from packaging import version
import requests
import json
from pe_automator._version import __version__
import os

def is_version_newer(project_json):
    # get the version URL from the project JSON
    if not os.path.exists(project_json):
        print(f"Project JSON file does not exist: {project_json}")
        return False
    try:
        with open(project_json, 'r') as f:
            project_data = json.load(f)
        version_url = project_data.get('version_url')
        if not version_url:
            print("Version URL not found in the project JSON.")
            return False
    except json.JSONDecodeError as e:
        print(f"Error decoding JSON: {e}")
        return False

    # Get the content of the version file
    try:
        response = requests.get(version_url)
        response.raise_for_status()  # Raise an error for HTTP errors
        min_version = response.text.strip()
    except requests.RequestException as e:
        print(f"Error fetching version file: {e}")
        return False
    
    is_new = version.parse(__version__.lstrip('v')) >= version.parse(min_version.lstrip('v'))

    print(f"Current version: {__version__}, Minimum required version: {min_version}")
    return is_new


# this is a command line script to extract the priors for GWTC-4 with given hdf5 files and save them to a file
import argparse
import h5py
import numpy as np
import re
import os
import shutil
import zipfile
import json


special_runs = {
    'GW230529_181500': 'C00:IMRPhenomXPHM:LowSpinSecondary'
}

default_run_name = 'C00:IMRPhenomXPHM-SpinTaylor'


class FlexibleKeyDict(dict):
    """
    A dictionary that allows flexible key access, replacing '-' with '_' and vice versa, to handle the different key formats
    used in bilby pipe
    """
    def get(self, key, default=None):
        if key in self:
            return super().get(key, default)

        # Try replacing - with _ and vice versa
        alt_key_1 = key.replace('-', '_')
        alt_key_2 = key.replace('_', '-')

        if alt_key_1 in self:
            return super().get(alt_key_1, default)
        elif alt_key_2 in self:
            return super().get(alt_key_2, default)

        return default
    

def get_gw_name_from_hdf5_filename(filename):
    match = re.search(r'GW\d{6}_\d{6}', filename)
    if match:
        return match.group(0)
    else:
        raise ValueError(f"Filename {filename} does not match expected format for GW name.")


def parse_dict_from_string(dict_string):
    dict_string = dict_string.strip('{} \n')
    # Split into key-value pairs
    dict_items = [item for item in dict_string.split(',') if item.strip()]
    # Build the dictionary
    parsed_dict = {}
    for item in dict_items:
        if ':' in item:
            key, value = item.split(':', 1)
            key = key.strip().strip('"').strip("'")
            value = value.strip().strip('"').strip("'")
           
            parsed_dict[key] = value

    return parsed_dict


def parse_array_from_string(array_string):
    if array_string.startswith('['):
        output_arr = eval(array_string)
    else:
        output_arr = [array_string]
    
    output_arr = [item.strip("'") for item in output_arr]
    
    return output_arr


def extract_prior(pe_samples, run_name):
    prior_keys = pe_samples[run_name]['priors']['analytic'].keys()

    prior = {}
    for key in prior_keys:
        value = pe_samples[run_name]['priors']['analytic'][key][0]
        if isinstance(value, bytes):
            value = value.decode('utf-8')
        prior[key] = value

    return prior


def extract_config(pe_samples, run_name):
    config_keys = pe_samples[run_name]['config_file']['config'].keys()

    config = FlexibleKeyDict({})
    for key in config_keys:
        value = pe_samples[run_name]['config_file']['config'][key][0]
        if isinstance(value, bytes):
            value = value.decode('utf-8')
        config[key] = value

    return config


def extract_info(hdf5_files, output_file):
    # create a tmp directory to store the priors
    os.makedirs('priors', exist_ok=True)
    os.makedirs('configs', exist_ok=True)
    os.makedirs('psds', exist_ok=True)
    os.makedirs('spline_cal_envs', exist_ok=True)
    os.makedirs('framefile_meta', exist_ok=True)

    for hdf5_file in hdf5_files:
        gw_name = get_gw_name_from_hdf5_filename(hdf5_file)
        
        with h5py.File(hdf5_file, 'r') as pe_samples:
            wf_run_name = special_runs.get(gw_name, default_run_name)

            # extract the prior for the given run name
            prior = extract_prior(pe_samples, wf_run_name)

            print(f"Extracting priors for {gw_name} with run name {wf_run_name}")
            with open(f'priors/{gw_name}_priors.ini', 'wb') as f:
                for key, value in prior.items():
                    f.write(f"{key} = {value}\n".encode('utf-8'))

            # extract the config for the given run name
            config = extract_config(pe_samples, wf_run_name)

            print(f"Extracting config for {gw_name} with run name {wf_run_name}")
            with open(f'configs/{gw_name}_config.ini', 'wb') as f:
                for key, value in config.items():
                    f.write(f"{key} = {value}\n".encode('utf-8'))

            psd_dict = parse_dict_from_string(config.get('psd-dict'))
            spline_cal_env_dict = parse_dict_from_string(config.get('spline-calibration-envelope-dict'))

            # copy the psd file to the psds directory
            for ifo, psd_file in psd_dict.items():
                dst_psd_path = os.path.join('psds', f"{gw_name}_{ifo}_psd.txt")
                shutil.copyfile(psd_file, dst_psd_path)

            # copy the spline calibration envelope file to the spline_cal_envs directory
            for ifo, spline_cal_env_file in spline_cal_env_dict.items():
                dst_spline_cal_env_path = os.path.join('spline_cal_envs', f"{gw_name}_{ifo}_spline_cal_env.txt")
                shutil.copyfile(spline_cal_env_file, dst_spline_cal_env_path)

            detectors = parse_array_from_string(config['detectors'])
            channel_dict = parse_dict_from_string(config.get('channel-dict'))
            trigger_time = float(config.get('trigger-time'))
            post_trigger_duration = float(config.get('post-trigger-duration', 0))
            duration = float(config['duration'])
            start_time = trigger_time + post_trigger_duration - duration
            end_time = start_time + duration
            required_data = {
                'gw_name': gw_name,
                'detectors': detectors,
                'channel_dict': channel_dict,
                'trigger_time': trigger_time,
                'post_trigger_duration': post_trigger_duration,
                'duration': duration,
                'start_time': start_time,
                'end_time': end_time
            }
            # save the required data to a json file
            with open(f'framefile_meta/{gw_name}.json', 'w') as f:
                json.dump(required_data, f, indent=4)


    # create a zip file with the priors and configs, keeping the directory structure
    with zipfile.ZipFile(output_file, 'w') as zipf:
        for dir in ['priors', 'configs', 'psds', 'spline_cal_envs', 'framefile_meta']:
            for root, dirs, files in os.walk(dir):
                for file in files:
                    zipf.write(os.path.join(root, file), os.path.relpath(os.path.join(root, file), os.path.join(dir, '..')))

    # clean up the tmp directory
    shutil.rmtree('priors')
    shutil.rmtree('configs')
    shutil.rmtree('psds')
    shutil.rmtree('spline_cal_envs')
    shutil.rmtree('framefile_meta')


def main():
    parser = argparse.ArgumentParser(description="Extract priors from GWTC-4 HDF5 files.")
    parser.add_argument('gwtc4_directory', type=str, help='Directory containing the GWTC-4 HDF5 files.')
    parser.add_argument('--output', type=str, default='gwtc4_info.zip',
                        help='Output file to save the extracted priors (default: gwtc4_info.zip)')
    

    args = parser.parse_args()

    # get all hdf5 files contains 'GW' in the gwtc4_directory
    hdf5_files = [os.path.join(args.gwtc4_directory, f) for f in os.listdir(args.gwtc4_directory) if f.endswith('.hdf5') and 'combined_PEDataRelease' in f]

    extract_info(hdf5_files, args.output)


if __name__ == "__main__":
    main()


from pe_automator.utils.io import read_samples
import numpy as np
from pesummary.utils.array import Array
from pe_automator.constants.approximants import PRECESSING_APPROXIMANTS, ECCENTRIC_APPROXIMANTS, PRECESSING_PARAMETERS, ECCENTRIC_PARAMETERS
from pesummary.gw.conversions import convert

def credible_intervals(samples):
    """
    Compute credible intervals for the given samples.
    """
    data = {}
    for event, models in samples.items():
        data[event] = {}
        for model, params in models.items():
            data[event][model] = {}
            for param, values in params.items():
                if values is None or len(values) == 0:
                    continue

                ci = values.credible_interval([5, 95])
                data[event][model][param] = {
                    "lower": ci[0],
                    "upper": ci[1],
                    "median": float(values.average("median"))
                }
    return data


def combine_samples(samples_list, parameters):
    combined_samples = {par: [] for par in parameters}
    all_wf_names = samples_list.keys()
    for par in parameters:
        if par in PRECESSING_PARAMETERS:
            wf_names = [wf for wf in all_wf_names if wf in PRECESSING_APPROXIMANTS]
        elif par in ECCENTRIC_PARAMETERS:
            wf_names = [wf for wf in all_wf_names if wf in ECCENTRIC_APPROXIMANTS]
        else:
            wf_names = list(all_wf_names)

        for wf in wf_names:
            combined_samples[par] = np.concatenate((combined_samples[par], samples_list[wf].get(par, [])))
        combined_samples[par] = Array(combined_samples[par])
    return combined_samples


def credible_intervals_from_data_dict(data_dict, parameters, eventnames):
    all_samples = {}

    # loop over eventname
    for eventname in eventnames:
        all_samples[eventname] = {}
        full_parameters = set(parameters)
        # loop over model/labels
        for model, file_path in data_dict[eventname].items():
            try:
                samples = read_samples(file_path, parameters)
                samples = {k: v for k, v in samples.items() if v.size > 0}
                all_samples[eventname][model] = convert(samples)

                if not hasattr(all_samples,'chi_1') and hasattr(all_samples,'spin_1z'):
                    all_samples[eventname][model]['chi_1'] = all_samples[eventname][model]['spin_1z']

                if not hasattr(all_samples,'chi_2') and hasattr(all_samples,'spin_2z'):
                    all_samples[eventname][model]['chi_2'] = all_samples[eventname][model]['spin_2z']

                full_parameters.update(all_samples[eventname][model].keys())
            except Exception as e:
                print(f"Error reading samples for {eventname} - {model}: {e}")
                raise e

        combined_samples = combine_samples(all_samples[eventname], full_parameters)
        all_samples[eventname]['Combined'] = combined_samples

    credible_intervals_dict = credible_intervals(all_samples)

    return credible_intervals_dict


from pe_automator.actions.gitlab import fetch_runs
import os
import json
import re
import bilby
import numpy as np


def get_jobs(access_token, gitlab_url, gitlab_project, approximant, status='all'):
    """
    Get the jobs for a given approximant and status.
    """
    runs = fetch_runs(gitlab_url=gitlab_url, gitlab_project=gitlab_project, private_token=access_token)

    if status == 'all':
        selected_runs = [r for r in runs['completed'] if r['approximant'] == approximant] + \
               [r for r in runs['ongoing'] if r['approximant'] == approximant]
    elif status == 'completed':
        selected_runs = [r for r in runs['completed'] if r['approximant'] == approximant]
    elif status == 'ongoing':
        selected_runs = [r for r in runs['ongoing'] if r['approximant'] == approximant]
    else:
        raise ValueError("Status must be 'all', 'completed', or 'ongoing'.")
    
    eventnames = [r['eventname'] for r in selected_runs]
    reformated_runs = { eventname: [] for eventname in eventnames }
    for run in selected_runs:
        reformated_runs[run['eventname']].append(run)

    return reformated_runs


def get_full_eventnames(data_folder):
    full_eventnames = []
    for filename in os.listdir(os.path.join(data_folder, 'configs')):
        if filename.endswith("_config.ini"):
            # extract the event name from the filename
            eventname = filename.split("_config.ini")[0]
            full_eventnames.append(eventname)

    # sort the event names
    full_eventnames.sort()
    return full_eventnames


def read_results_catalog(results_folder='results', results_catalog_file='results_catalog.json'):
    """
    Read the results catalog from a JSON file.
    """
    return Catalog(os.path.join(results_folder, results_catalog_file))


def get_gwtc_hdf5_filename(gwtc4_directory, run_name='C00:IMRPhenomXPHM-SpinTaylor', special_runs={}):
    """
    Functions to get the GWTC HDF5 filename for a given event and run name.

    Arguments:
    - gwtc4_directory: The directory containing the GWTC-4 HDF5 files.
    - run_name: The name of the run to use if no special run is found.
    - special_runs: A dictionary mapping event names to special run names.

    """

    hdf5_files = [os.path.join(gwtc4_directory, f) for f in os.listdir(gwtc4_directory) if f.endswith('.hdf5') and 'combined_PEDataRelease' in f]

    gwtc_files = {}
    for filename in hdf5_files:
        match = re.search(r'GW\d{6}_\d{6}', filename)
        if match:
            eventname = match.group(0)
            gwtc_files[eventname] = {
                'filename': filename,
                'run_name': special_runs.get(eventname, run_name),
                'merged': True
            }
        else:
            print(f"Warning: No GWTC-4 match found in filename {filename}")

    return gwtc_files


class Catalog(dict):
    """
    A dictionary subclass to hold catalog information.
    """
    def __init__(self, filename):
        super().__init__()
        self.read(filename)

    def read(self, filename):
        with open(filename, 'r') as f:
            data = json.load(f)
        self.update(data)
        self.__dict__ = self

    def get_run(self, eventname, approx, label):
        runs = self.get(eventname, [])
        for run in runs:
            if run['approximant'] == approx and run['run_label'] == label:
                return run
        return None
    
    def get_result_file_base(self, eventname, approx, label):
        run = self.get_run(eventname, approx, label)
        if run:
            if '_data0_' in run['file_path']:
                return run['file_path'].split('/')[2].rsplit('_data0_', 1)[0]
            else:
                raise ValueError(f"Unexpected file_path format: {run['file_path']}")
        return None

    def get_result(self, eventname, approx, label):
        run = self.get_run(eventname, approx, label)
        if run:
            return bilby.result.read_in_result(filename=run['file_path'])
        return None

    def get_posterior(self, eventname, approx, label):
        run = self.get_run(eventname, approx, label)
        if run:
            return bilby.result.read_in_result(filename=run['file_path']).posterior
        return None

    def get_config_path(self, eventname, approx, label):
        run = self.get_run(eventname, approx, label)
        file_base = self.get_result_file_base(eventname, approx, label)
        if run:
            return f'results/{run["eventname"]}/{file_base}_config_complete.ini'
        return None
    
    def get_recomputed_likelihood_path(self, data_dir, eventname, approx, label):
        return os.path.join(data_dir, f"{eventname}_{approx}_{label}_recomputed_likelihood.dat")
    
    def get_recomputed_likelihood(self, data_dir, eventname, approx, label):
        file_path = self.get_recomputed_likelihood_path(data_dir, eventname, approx, label)
        if os.path.exists(file_path):
            return np.loadtxt(file_path)
        return None



from datetime import datetime
import os
import json
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
from matplotlib.patches import Patch
from matplotlib.lines import Line2D
from itertools import combinations
from pesummary.utils.utils import jensen_shannon_divergence_from_samples, jensen_shannon_divergence_from_pdfs
from .read import read_samples
from ..constants.events import (
    eccentricCandidates,
    glitchedEvents,
    problemEvents,
    waveformSystematicsLVK,
    glitchedeccentricCandidates,
    eccentricSystematics,
    eccentricAEISystematics,
    glitchedSystematics,
    multimodalities,
    newSystematics
)

model_colors = {
    'GWTC': 'grey',
    'IMRPhenomXPNR': 'royalblue',
    'IMRPhenomXE': 'orchid',
    'IMRPhenomTEHM': 'forestgreen',
    'IMRPhenomTHM_20': 'crimson',
    'IMRPhenomTHM': 'orange',
    'IMRPhenomTPHM': 'gold',
    'IMRPhenomXPHM': 'lightseagreen'
}

event_group_colors = {
    'eccentricCandidates': 'orange',
    'glitchedEvents': 'dodgerblue',
    'problemEvents': 'black',
    'waveformSystematicsLVK': 'green',
    'glitchedeccentricCandidates': 'crimson',
    'eccentricSystematics': 'darkturquoise',
    'eccentricAEISystematics': 'purple',
    'glitchedSystematics': 'peru',
    'multimodalities': 'mediumseagreen',
    'newSystematics': 'orchid'
}

def get_event_color(event):
    # if event in eccentricCandidateAEI:
    #     return event_group_colors['eccentricCandidateAEI']
    if event in eccentricCandidates:
        return event_group_colors['eccentricCandidates']
    elif event in glitchedEvents:
        return event_group_colors['glitchedEvents']
    elif event in problemEvents:
        return event_group_colors['problemEvents']
    elif event in waveformSystematicsLVK:
        return event_group_colors['waveformSystematicsLVK']
    elif event in glitchedeccentricCandidates:
        return event_group_colors['glitchedeccentricCandidates']
    elif event in eccentricSystematics:
        return event_group_colors['eccentricSystematics']
    elif event in eccentricAEISystematics:
        return event_group_colors['eccentricAEISystematics']
    elif event in glitchedSystematics:
        return event_group_colors['glitchedSystematics']
    elif event in multimodalities:
        return event_group_colors['multimodalities']
    elif event in newSystematics:
        return event_group_colors['newSystematics']
    else:
        return 'gray' 

def posteriors_1d(pe_file_dict, parameters, filename, eventnames=None):
    if eventnames is None:
        selected_pe_file_dict = dict(sorted(list(pe_file_dict.items())))
    else:
        selected_pe_file_dict = {event: pe_file_dict[event] for event in eventnames if event in pe_file_dict}

    n_events = len(selected_pe_file_dict)
    n_params = len(parameters)
    fig, axes = plt.subplots(nrows=n_events, ncols=n_params, figsize=(3*n_params, 2.5*n_events), sharex=False, sharey=False)
    fig.subplots_adjust(wspace=0.3, hspace=0.3)
    now = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    plt.suptitle(f'dlogz Progress Check at {now}', fontsize=16)

    if n_events == 1:
        axes = np.expand_dims(axes, axis=0)
    if n_params == 1:
        axes = np.expand_dims(axes, axis=1)

    skip_dict = {
        'IMRPhenomTHM': ['tilt_1', 'tilt_2', 'phi_12', 'phi_jl', 'chi_p', 'eccentricity', 'mean_anomaly'],
        'IMRPhenomXPNR': ['eccentricity', 'mean_anomaly'],
        'IMRPhenomXPHM': ['eccentricity', 'mean_anomaly'],
        'IMRPhenomTPHM': ['eccentricity', 'mean_anomaly'],
        'IMRPhenomTEHM': ['tilt_1', 'tilt_2', 'phi_12', 'phi_jl', 'chi_p'],
        'IMRPhenomTHM_20': ['tilt_1', 'tilt_2', 'phi_12', 'phi_jl', 'chi_p', 'eccentricity', 'mean_anomaly'],
        'IMRPhenomXE': ['tilt_1', 'tilt_2', 'phi_12', 'phi_jl', 'chi_p'],
    }

    for i, (event, model_files) in enumerate(selected_pe_file_dict.items()):
        for j, (param_key, param_label) in enumerate(parameters.items()):
            ax = axes[i, j]
            ax.set_yticks([])
            for model in model_colors:
                if model in model_files:
                    color = model_colors[model]
                    if model in skip_dict and param_key in skip_dict[model]:
                        continue
                    try:
                        samples = read_samples(model_files[model], param_key)
                        ax.hist(samples, bins=60, density=True, histtype='step', lw=1.5,
                                label=model if i == 0 and j == 0 else "", color=color)
                    except Exception as e:
                        print(f"Missing {param_key} for {event} in {model}: {e}")
                        continue

            ax.set_xlabel(param_label, fontsize=16)

            if j == 0:
                event_color = get_event_color(event)
                event_group_label = None

                for group_name, event_list in {
                    'Eccentric Candidate': eccentricCandidates,
                    'Glitched Event': glitchedEvents,
                    'Problematic Event': problemEvents,
                    'Waveform Systematics LVK': waveformSystematicsLVK,
                    'Glitched + Eccentric Candidate': glitchedeccentricCandidates,
                    'Eccentric Candidate + Waveform Systematics LVK': eccentricSystematics,
                    'EccentricAEI + Waveform Systematics LVK': eccentricAEISystematics,
                    'Glitched Event + Waveform Systematics LVK': glitchedSystematics,
                    'Multimodalities': multimodalities,
                }.items():
                    if event in event_list:
                        event_group_label = group_name
                        break

                ax.set_ylabel(event, fontsize=14, color=event_color)

                if event_group_label is not None:
                    ax.set_title(f"{event_group_label}", fontsize=16, loc='left', pad=10, color=event_color)


    handles_models = [Patch(facecolor='none', edgecolor=model_colors[model], linewidth=2) for model in model_colors]
    labels_models = [model for model in model_colors]

    plt.tight_layout(rect=[0, 0, 1, 0.9])

    leg1 = fig.legend(
        handles_models, labels_models,
        loc='upper left', bbox_to_anchor=(0, 1.003),
        ncol=5, frameon=True, fontsize=16
    )

    fig.add_artist(leg1)

    plt.tight_layout(rect=[0, 0, 1, 1])
    if filename:
        plt.savefig(filename, transparent=True, bbox_inches='tight')
    else:
        plt.show()


def js_divergence_plot(pe_file_dict, parameters, outdir, eventnames=None, model_pairs=None):
    if eventnames is None:
        selected_pe_file_dict = pe_file_dict
    else:
        selected_pe_file_dict = {event: pe_file_dict[event] for event in eventnames if event in pe_file_dict}

    if model_pairs is None:
        all_models = set()
        for model_dict in selected_pe_file_dict.values():
            all_models.update(model_dict.keys())
        model_pairs = list(combinations(sorted(all_models), 2))

    color_list = ['royalblue', 'orange', 'forestgreen', 'crimson', 'purple', 'teal', 'brown', 'gold']
    marker_list = ['o', 's', 'D', '^', 'v', 'P', '*', 'X']

    all_jsd = {}
    all_comparisons = set()
    event_labels = []

    for event_idx, (event, model_dict) in enumerate(reversed(list(selected_pe_file_dict.items()))):
        event_labels.append(event)

        samples = {}
        for model, path in model_dict.items():
            try:
                samples[model] = read_samples(path, parameters)
            except Exception as e:
                print(f"Error reading samples for {event} {model} in file {path}: {e}")
                raise e

        for model1, model2 in model_pairs:
            if model1 not in samples or model2 not in samples:
                continue

            comp_key = f"{model1} vs {model2}"
            all_comparisons.add(comp_key)
            if comp_key not in all_jsd:
                all_jsd[comp_key] = {param: [np.nan]*event_idx for param in parameters}

            for param in parameters:
                s0 = samples[model1].get(param)
                s1 = samples[model2].get(param)

                if s0 is None or s1 is None:
                    js = np.nan
                else:
                    try:
                        js = jensen_shannon_divergence_from_samples([s0, s1])
                    except Exception as e:
                        print(f"{event} {comp_key} {param}: Error computing JSD – {e}")
                        js = np.nan

                while len(all_jsd[comp_key][param]) < event_idx:
                    all_jsd[comp_key][param].append(np.nan)
                all_jsd[comp_key][param].append(js)

    n_events = len(event_labels)
    for comp in all_comparisons:
        for param in parameters:
            while len(all_jsd[comp][param]) < n_events:
                all_jsd[comp][param].append(np.nan)

    comparisons = sorted(all_comparisons)
    colors = {comp: color_list[i % len(color_list)] for i, comp in enumerate(comparisons)}
    markers = {comp: marker_list[i % len(marker_list)] for i, comp in enumerate(comparisons)}

    n_params = len(parameters)
    fig, axes = plt.subplots(2, n_params, figsize=(3.5*n_params, 0.5*n_events + 4), gridspec_kw={"height_ratios": [1.5, n_events]}, sharex='col')

    for i, param in enumerate(parameters):
        ax_hist = axes[0, i]
        ax_main = axes[1, i]
        ax_main.set_yticks(np.arange(n_events))
        
        yticks = np.arange(n_events)
        yticklabels = event_labels
        ax_main.set_yticklabels(yticklabels if i == 0 else [])
        if i == 0:
            for tick_label, event in zip(ax_main.get_yticklabels(), event_labels):
                tick_label.set_color(get_event_color(event))

        ax_main.set_xscale("log")
        ax_hist.set_xscale("log")

        for comp in comparisons:
            js_values = np.array(all_jsd[comp][param])
            y = np.arange(n_events)
            x = js_values
            ax_main.scatter(x, y, label=comp if i == 0 else "", color=colors[comp], marker=markers[comp], s=60, alpha=0.6)

            valid = x[~np.isnan(x)]
            if valid.size > 0:
                ax_hist.hist(valid, bins=np.logspace(np.log10(valid.min()), np.log10(valid.max()), 20), histtype="step", color=colors[comp], lw=1.5, density=False)

        ax_main.axvline(0.02, color='darkred', linestyle='--', linewidth=1)
        ax_hist.axvline(0.02, color='darkred', linestyle='--', linewidth=1)
        ax_main.set_ylim(-1, n_events)
        ax_main.grid(True, axis='x', linestyle=':', alpha=0.4)

    for ax, param in zip(axes[1], parameters):
        ax.set_xlabel(parameters[param] + r" [JSD (bits)]", fontsize=14)

    handles = [Line2D([], [], marker=markers[comp], linestyle='None', color=colors[comp], markersize=10, label=comp, alpha=0.6)
        for comp in comparisons]
    handles_event_groups = [Patch(color=color, label=group)
                            for group, color in event_group_colors.items()]

    leg1 = fig.legend(
        handles, [h.get_label() for h in handles],
        loc='upper left', bbox_to_anchor=(0, 1.016),
        ncol=3, frameon=True, fontsize=16)
    leg2 = fig.legend(handles_event_groups, list(event_group_colors.keys()),
                      loc='upper left', bbox_to_anchor=(0, 1.), ncol=4,
                      frameon=True, fontsize=16)
    
    fig.add_artist(leg1)
    fig.add_artist(leg2)
    
    plt.tight_layout(rect=[0, 0, 1, 0.98])
    if outdir:
        plt.savefig(outdir+'/js-divergence.png', transparent=True, bbox_inches='tight')
    else:
        plt.show()


def event_posterior_plot(event, model_files, parameters, axes):
    skip_dict = {
        'IMRPhenomTHM': ['tilt_1', 'tilt_2', 'phi_12', 'phi_jl', 'chi_p', 'eccentricity', 'mean_anomaly'],
        'IMRPhenomXPNR': ['eccentricity', 'mean_anomaly'],
        'IMRPhenomXPHM': ['eccentricity', 'mean_anomaly'],
        'IMRPhenomTPHM': ['eccentricity', 'mean_anomaly'],
        'IMRPhenomTEHM': ['tilt_1', 'tilt_2', 'phi_12', 'phi_jl', 'chi_p'],
        'IMRPhenomTHM_20': ['tilt_1', 'tilt_2', 'phi_12', 'phi_jl', 'chi_p', 'eccentricity', 'mean_anomaly'],
        'IMRPhenomXE': ['tilt_1', 'tilt_2', 'phi_12', 'phi_jl', 'chi_p'],
    }

    for j, (param_key, param_label) in enumerate(parameters.items()):
        ax = axes[j]
        ax.set_yticks([])
        for model in model_colors:
            if model in model_files:
                color = model_colors[model]
                if model in skip_dict and param_key in skip_dict[model]:
                    continue
                try:
                    samples = read_samples(model_files[model], param_key)
                    ax.hist(samples, bins=60, density=True, histtype='step', lw=1.5,
                            label=model if j == 0 else "", color=color)
                except Exception as e:
                    print(f"Missing {param_key} for {event} in {model}: {e}")
                    continue

        ax.set_xlabel(param_label, fontsize=16)
        if j == 0:
            event_color = get_event_color(event)
            event_group_label = None

            for group_name, event_list in {
                'Eccentric Candidate': eccentricCandidates,
                'Glitched Event': glitchedEvents,
                'Problematic Event': problemEvents,
                'Waveform Systematics LVK': waveformSystematicsLVK,
                'Glitched + Eccentric Candidate': glitchedeccentricCandidates,
                'Eccentric Candidate + Waveform Systematics LVK': eccentricSystematics,
                'EccentricAEI + Waveform Systematics LVK': eccentricAEISystematics,
                'Glitched Event + Waveform Systematics LVK': glitchedSystematics,
                'Multimodalities': multimodalities,
            }.items():
                if event in event_list:
                    event_group_label = group_name
                    break

            ax.set_ylabel(event, fontsize=14, color=event_color)

            if event_group_label is not None:
                ax.set_title(f"{event_group_label}", fontsize=16, loc='left', pad=10, color=event_color)


def catalog_posterior_plot(pe_file_dict, parameters, outdir, eventnames=None):
    n_params = len(parameters)
    event_data = {}
    for event, model_files in pe_file_dict.items():
        fig, axes = plt.subplots(nrows=1, ncols=n_params, figsize=(3*n_params, 2.5*1), sharex=False, sharey=False)
        fig.subplots_adjust(wspace=0.3, hspace=0.3)
        event_posterior_plot(event, model_files, parameters, axes)
        plt.tight_layout()
        plt.savefig(os.path.join(outdir, f"plots/{event}_posteriors.png"), transparent=True, bbox_inches='tight')
        plt.close()

        event_group_label = []
        for group_name, event_list in {
            'Eccentric Candidate': eccentricCandidates,
            'Glitched Event': glitchedEvents,
            'Problematic Event': problemEvents,
            'Waveform Systematics LVK': waveformSystematicsLVK,
        }.items():
            if event in event_list:
                event_group_label.append(group_name)
        event_data[event] = {
            'labels': event_group_label,
            'eccentricity': {
                'median': 0.1,
            },
            'total_mass': {
                'median': 50,
            },
            'mass_ratio': {
                'median': 0.5,
            },
            'snr_H1': {
                'median': 100,
            },
            'snr_L1': {
                'median': 100,
            },
            'snr_network': {
                'median': 100,
            },
        }

        html_str = plot_page(event_data)
        with open(os.path.join(outdir, f"plots/{event}_posteriors.html"), 'w') as f:
            f.write(html_str)


def plot_page(data_dict):
    template = """
<!DOCTYPE html>
<html lang="en">
<head>
<meta charset="UTF-8">
<title>Event Viewer</title>
<style>
    body { font-family: Arial, sans-serif; margin: 20px; }
    table { border-collapse: collapse; width: 100%; }
    th, td { border: 1px solid #ccc; padding: 8px; text-align: left; vertical-align: top; }
    th { background-color: #f4f4f4; }
    .label-badge { 
        background-color: #eee; 
        border-radius: 4px; 
        padding: 2px 6px; 
        margin: 0 4px; 
        font-size: 0.85em;
    }
    .plot-row td {
        text-align: center;
        background-color: #fafafa;
    }
    .plot-row img {
        max-width: 100%;
        height: auto;
    }
    .sortable { cursor: pointer; color: blue; text-decoration: underline; }
    .filters { margin-bottom: 15px; }
</style>
</head>
<body>

<h1>Event Viewer</h1>

<div class="filters">
    <label for="labelFilter">Filter by Label:</label>
    <select id="labelFilter">
        <option value="">-- All --</option>
    </select>
</div>

<table id="eventsTable">
    <thead>
        <tr>
            <th>Event</th>
            <th>Labels</th>
            <th class="sortable" onclick="sortTable('eccentricity')">Eccentricity (median)</th>
            <th class="sortable" onclick="sortTable('total_mass')">Total Mass (median)</th>
            <th class="sortable" onclick="sortTable('mass_ratio')">Mass Ratio (median)</th>
            <th class="sortable" onclick="sortTable('snr_H1')">SNR H1 (median)</th>
            <th class="sortable" onclick="sortTable('snr_L1')">SNR L1 (median)</th>
            <th class="sortable" onclick="sortTable('snr_network')">SNR Network (median)</th>
        </tr>
    </thead>
    <tbody></tbody>
</table>

<script>
const data = {};

let sortDirection = 1;

function populateTable() {
    const tbody = document.querySelector("#eventsTable tbody");
    tbody.innerHTML = "";
    const selectedLabel = document.getElementById("labelFilter").value;

    for (const [name, event] of Object.entries(data)) {
        if (selectedLabel && !event.labels.includes(selectedLabel)) continue;

        // Metadata row
        const metaRow = document.createElement("tr");
        metaRow.innerHTML = `
            <td>${name}</td>
            <td>${event.labels.map(l => `<span class="label-badge">${l}</span>`).join("")}</td>
            <td>${event.eccentricity.median}</td>
            <td>${event.total_mass.median}</td>
            <td>${event.mass_ratio.median}</td>
            <td>${event.snr_H1.median}</td>
            <td>${event.snr_L1.median}</td>
            <td>${event.snr_network.median}</td>
        `;
        tbody.appendChild(metaRow);

        // Plot row
        const plotRow = document.createElement("tr");
        plotRow.className = "plot-row";
        const plotCell = document.createElement("td");
        plotCell.colSpan = 8;
        plotCell.innerHTML = `<img src="${name}_posteriors.png" alt="${name}_posteriors plot">`;
        plotRow.appendChild(plotCell);
        tbody.appendChild(plotRow);
    }
}

function populateLabelFilter() {
    const select = document.getElementById("labelFilter");
    const labels = new Set();
    for (const event of Object.values(data)) {
        event.labels.forEach(l => labels.add(l));
    }
    for (const label of labels) {
        const option = document.createElement("option");
        option.value = label;
        option.textContent = label;
        select.appendChild(option);
    }
}

function sortTable(param) {
    const entries = Object.entries(data);
    entries.sort((a, b) => sortDirection * (a[1][param].median - b[1][param].median));
    sortDirection *= -1;
    const sortedData = {};
    for (const [name, event] of entries) sortedData[name] = event;
    Object.assign(data, sortedData);
    populateTable();
}

document.getElementById("labelFilter").addEventListener("change", populateTable);

populateLabelFilter();
populateTable();
</script>

</body>
</html>"""
    return template.replace("data = {};", f"data = {json.dumps(data_dict, indent=4)};")


import json
import shutil
from pathlib import Path
from ..actions.remote import run_command, upload_files, download_files
from ..actions.allocation import get_all_allocations
from filecmp import cmp

def setup_env(env_name, source_env, source_remote, data_dir):
    # download the packed environment
    print(f"Downloading packed environment {env_name} from {source_remote}:{source_env}")
    # This is a placeholder for the actual download logic
    ext = ''.join(Path(source_env).suffixes)
    env_filename = f"{env_name}{ext}"
    packed_env = f"{data_dir}/envs/{env_filename}"
    if source_remote:
        download_files(source_env, packed_env, source_remote, is_file=True)
    else:
        # skip if two files are the same
        if Path(source_env).exists() and Path(packed_env).exists() and cmp(source_env, packed_env, shallow=False):
            print(f"Environment {env_name} already exists in {data_dir}/envs, skipping download")
        else:
            print(f"Copying environment {env_name} to {data_dir}/envs")
            Path(data_dir, 'envs').mkdir(parents=True, exist_ok=True)
            shutil.copy(source_env, packed_env)

    # get the allocations
    allocations = get_all_allocations(data_dir)

    # upload the environment to the remote servers
    for name, allocation in allocations.items():
        hostname = allocation.get('hostname')
        maintainer = allocation.get('maintainer')
        scratch = allocation.get('scratch')
        
        if not (hostname and maintainer and scratch):
            print(f"No remote specified for allocation {name}, skipping upload")

        remote = f"{maintainer}@{hostname}"
        env_path = f"{scratch}/{maintainer}/envs"
        target_path = f"{env_path}/{env_name}"

        # check if the conda env exists on the remote server by checking the target path is not empty
        check_command = f"if [ -d {target_path} ] && [ \"$(ls -A {target_path})\" ]; then echo 'exists'; else echo 'not exists'; fi"
        exists = run_command(check_command, remote)
        if exists.strip() == "exists":
            print(f"Environment {env_name} already exists on {remote}, skipping upload")
        else:
            # upload the environment
            print(f"Uploading environment {env_name} to {remote}:{target_path}")
            # create the target directory if it does not exist
            create_dir_command = f"mkdir -p {env_path}"
            run_command(create_dir_command, remote)
            upload_files(packed_env, env_path, remote)

            # set the conda environment
            print(f"Unpacking environment {env_name} on {remote}:{target_path}")
            set_env_command = f"mkdir -p {target_path} && tar -xzf {env_path}/{env_filename} -C {target_path}"
            run_command(set_env_command, remote)

            # unpack the environment
            # print(f"Unpacking environment {env_name} on {remote}:{target_path}")
            # unpack_command = f"source {target_path}/bin/activate && conda-unpack "
            # run_command(unpack_command, remote)

        # check if the environment is set correctly
        print(f"Checking if environment {env_name} is set correctly on {remote}")
        # try import bilby
        check_command = f"source {target_path}/bin/activate && python -c 'import bilby; print(bilby.__version__)'"
        version = run_command(check_command, remote)
        if version:
            print(f"Environment {env_name} is set correctly on {remote}, version: {version.strip()}")


#!/usr/bin/env python

"""
This is an example script to run PESummary from within a python script for
non-gravitational wave data.
"""
from pesummary.core import command_line, inputs, finish
from pesummary.core.file import meta_file
from pesummary.cli.summarypages import main


def generate_summary_page(run_info):
    """
    Generate a summary page for the given run information.

    Parameters
    ----------
    run_info : dict
        A dictionary containing the run information, including 'file_path',
        'eventname', and 'title'.
    """

    from pesummary.cli.summarypages import main

    main_args = [
        "--samples", run_info['file_path'],
        "--webdir", f'pesummary/{run_info["eventname"]}',
        "--label", run_info['title'],
        "--gw", "true"
    ]
    
    main(main_args)




# simple_ssh_client.py
import paramiko
import threading

class SimpleSSHClient:
    def __init__(self, name, hostname, username, password=None, port=22, key_filename=None, timeout=10):
        self.name = name
        self.hostname = hostname
        self.port = port
        self.username = username
        self.password = password
        self.key_filename = key_filename
        self.timeout = timeout
        self.client = None
        self.lock = threading.Lock()

    def connect(self):
        if self.client is None or not self.client.get_transport().is_active():
            self.client = paramiko.SSHClient()
            self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            if self.key_filename or self.password:
                self.client.connect(
                    hostname=self.hostname,
                    port=self.port,
                    username=self.username,
                    password=self.password,
                    key_filename=self.key_filename,
                    timeout=self.timeout
                )
            else:
                agent = paramiko.Agent()
                keys = agent.get_keys()
                if not keys:
                    raise ValueError("No SSH keys found in the agent.")
                for key in keys:
                    try:
                        self.client.connect(
                            hostname=self.hostname,
                            port=self.port,
                            username=self.username,
                            pkey=key,
                            timeout=self.timeout
                        )
                        break  # If connection is successful, exit the loop
                    except paramiko.AuthenticationException:
                        continue

    def execute(self, command, get_output=True):
        with self.lock:
            if self.client is None:
                raise Exception("SSH client not connected. Call connect() first.")

            stdin, stdout, stderr = self.client.exec_command(command)
            if get_output:
                out = stdout.read().decode()
                err = stderr.read().decode()
                return out.strip(), err.strip()
            return None, None

    def close(self):
        if self.client:
            self.client.close()
            self.client = None

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


class MultiSSHManager:
    def __init__(self, server_configs = []):
        """
        server_configs: List of dicts with keys:
            - hostname, username, password or key_filename, port (optional)
        """
        self.clients = {}
        self.lock = threading.Lock()

        for config in server_configs:
            name = config['name']
            self.clients[name] = SimpleSSHClient(**config)

    def add_client(self, name, hostname, username, password=None, port=22, key_filename=None, timeout=10, skip_if_exists=False):
        if name in self.clients:
            if skip_if_exists:
                return
            raise ValueError(f"Client with name {name} already exists.")
        self.clients[name] = SimpleSSHClient(name, hostname, username, password, port, key_filename, timeout)
        self.clients[name].connect()

    def get_client(self, name):
        if name not in self.clients:
            raise KeyError(f"No such host: {name}")
        return self.clients[name]

    def is_client_connected(self, name):
        if name not in self.clients:
            raise KeyError(f"No such host: {name}")
        return self.clients[name].client is not None and self.clients[name].client.get_transport().is_active()

    def connect_all(self):
        threads = []
        for name, client in self.clients.items():
            t = threading.Thread(target=client.connect)
            t.start()
            threads.append(t)
        for t in threads:
            t.join()

    def execute_on(self, name, command):
        """
        Returns: (stdout, stderr) or raises KeyError
        """
        if name not in self.clients:
            raise KeyError(f"No such host: {name}")
        return self.clients[name].execute(command)

    def close_all(self):
        for client in self.clients.values():
            client.close()

    def __enter__(self):
        self.connect_all()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close_all()


import re
from datetime import datetime
from pe_automator.actions.remote import run_command
from pe_automator.utils.parse_ssh_output import clean_env_in_string


def check_data_generated(run_info, ssh_client):
    """
    Check if the data generation has been completed for the given run information.
    
    Parameters:
    - run_info: Dictionary containing run information including 'run_dir' and 'remote'.
    
    Returns:
    - bool: True if data generation is complete, False otherwise.
    """
    
    run_dir = run_info['run_dir']
    
    # Command to check the last line of the error log file
    command = f'ls {run_dir}/log_data_generation/*.err 1> /dev/null 2>&1 && tail -n 1 {run_dir}/log_data_generation/*.err || echo "NO_ERR_FILES_FOUND"'
    
    # Execute the command on the remote server
    # remote = f"{user}@{run_info['hostname']}"
    # result = run_command(command, remote, silent=True)
    result, err = ssh_client.execute(command, get_output=True)

    if result.strip() == "NO_ERR_FILES_FOUND":
        return 'pending', ''
    
    if "Completed data generation" in result:
        # This means the data generation has completed successfully and moving the next stage. So no status should be recorded.
        return '', ''
    elif "Error" in result:
        # If there is an error in the log, return the error message
        return 'failed', result
    else:
        # If the log does not indicate completion or error, assume it is still running
        return 'pending', ''



def get_data_analysis_log(run_info, ssh_client):
    """
    Get the last line of the data analysis log for the given run information.
    
    Parameters:
    - run_info: Dictionary containing run information including 'run_dir' and 'remote'.
    
    Returns:
    - str: The last line of the data analysis log.
    """
    # remote = f"{user}@{run_info['hostname']}"
    run_dir = run_info['run_dir']

    # Execute the command on the remote server
    log_files_str, err = ssh_client.execute(f'bash -c "ls {run_dir}/log_data_analysis/*.err 1> /dev/null 2>&1 && stat --format=\'%Y %n\' {run_dir}/log_data_analysis/*  || echo \"NO_ERR_FILES_FOUND\""')
    log_files_str = clean_env_in_string(log_files_str)  # Clean the output to remove unwanted lines

    if log_files_str.strip() == "NO_ERR_FILES_FOUND":
        return []
    files = []

    for line in log_files_str.split('\n'):
        if not line.strip():
            continue
        parts = line.split()
        timestamp = int(parts[0])
        filename = ' '.join(parts[1:])
        file_info = {
            'timestamp': timestamp,
            'filename': filename,
            'datetime': datetime.fromtimestamp(timestamp).strftime('%Y-%m-%d %H:%M:%S')
        }

        if filename.endswith('.err'):
            file_info['type'] = 'err'
        elif filename.endswith('.out'):
            file_info['type'] = 'out'

        if re.search(r'par\d+\.(err|out)', filename):
            file_info['job_type'] = 'analysis'
        elif re.search(r'merge.(err|out)', filename):
            file_info['job_type'] = 'merge'
        elif re.search(r'merge_plot.(err|out)', filename):
            file_info['job_type'] = 'merge_plot'
        elif re.search(r'merge_final_result.(err|out)', filename):
            file_info['job_type'] = 'merge_final_result'
        
        timestamp_now = int(datetime.now().timestamp())
        file_info['recent'] = (timestamp_now - timestamp) < 60 * 30
        files.append(file_info)

    return files


def check_data_analysis(run_info, ssh_client):
    """
    Check if the data analysis has been completed for the given run information.
    
    Parameters:
    - run_info: Dictionary containing run information including 'run_dir' and 'remote'.
    
    Returns:
    - bool: True if data analysis is complete, False otherwise.
    """
    status = None

    files = get_data_analysis_log(run_info, ssh_client)

    if not files:
        # this means the generation is done but no analysis has been started yet
        return 'pending', ''

    merge_logs = [f for f in files if f['job_type'] == 'merge' and f['type'] == 'err']

    # if reached the merge stage, check the last merge log
    if len(merge_logs) > 0:
        merge_log = merge_logs[-1]

        result, err = ssh_client.execute(f"tail -n 10 {merge_log['filename']}", get_output=True)

        # if 'Combined results have \d+\ samples' in result:
        if re.search(r'Combined results have \d+ samples', result):
            status = 'completed'
            return status, ''
        else:
            if merge_log['recent']:
                status = 'merging'
                return status, ''
            else:
                status = 'failed'
                return status, result

    # if not reached the merge stage, check the analysis logs
    # first check if any partial analysis logs finished      
    partial_err_logs = [f for f in files if f['job_type'] == 'analysis' and f['type'] == 'err']
    is_analysis_completed = []
    is_analysis_failed = []
    messages = []
    for log in partial_err_logs:
        result, err = ssh_client.execute(f"tail -n 10 {log['filename']}", get_output=True)
  
        if re.search(r'Run completed', result):
            is_analysis_completed.append(True)
        else:
            is_analysis_completed.append(False)
            if re.search(r'Error', result):
                is_analysis_failed.append(True)
            else:
                is_analysis_failed.append(False)
        messages.append(result)
    if is_analysis_completed and all(is_analysis_completed):
        status = 'analysis_completed'
        return status, ''
    if any(is_analysis_failed):
        status = 'failed'
        return status, messages
    
    # TODO: if not failed nor completed, check if any analysis logs are still running
    if any(f['recent'] for f in files if f['job_type'] == 'analysis'):
        # if any analysis logs are recent, assume it is still running
        status = 'running'
        return status, ''
    # if no analysis logs are recent, assume it is might be hung
    else:
        # check if there are any recent logs
        if any(f['recent'] for f in files):
            status = 'running'
            return status, ''
        else:
            # no recent logs, assume it is failed
            status = 'hold'
            return status, ''
    



import requests
import argparse
import os


def get_prefered_pe(event_name, detail_url):
    response = requests.get(detail_url)
    if response.status_code == 200:
        data = response.json()
    else:
        raise Exception(f"Failed to fetch event info: {response.status_code} - {response.text}")
    is_GWTC1 = data.get('catalog').startswith('GWTC-1-confident')
    parameters_url = data.get('parameters_url', None)
    response = requests.get(parameters_url + '?pagesize=1000')
    if response.status_code == 200:
        event_data = response.json()
    else:
        raise Exception(f"Failed to fetch event parameters: {response.status_code} - {response.text}")
    if event_data['num_pages'] > 1:
        raise NotImplementedError("Pagination not implemented yet for event listing.")
    
    if not event_data:
        raise ValueError(f"No data found for event {event_name} with parameters URL {parameters_url}")

    results = event_data.get('results', [])

    # get the combined PE (labelled as 'preferred' in the gwosc) and the XPHM for extracting config
    gwosc_preferred_pe = None

    for result in results:
        if is_GWTC1 and result.get('name', '').startswith('GWTC-1-confident') and result.get('name', '').endswith('R2_pe_combined'):
            gwosc_preferred_pe = result
            break

        if result.get('is_preferred', False):
            gwosc_preferred_pe = result
            break

    return gwosc_preferred_pe


def get_all_event(catalog = 'GWTC'):
    url = f"https://gwosc.org/api/v2/catalogs/{catalog}/events?pagesize=1000"
    print(f"Fetching event list from {url}")
    response = requests.get(url)
    if response.status_code == 200:
        data = response.json()
    else:
        raise Exception(f"Failed to fetch event info: {response.status_code} - {response.text}")    
    
    if data['num_pages'] > 1:
        raise NotImplementedError("Pagination not implemented yet for event listing.")
    print(f"Found {len(data['results'])} events in catalog {catalog}")
    # return sort by name in data['results']
    return sorted(data['results'], key=lambda x: x['name'])


def download_file(url, output_path):
    # check if the file already exists
    if os.path.exists(output_path):
        print(f"File {output_path} already exists, skipping download.")
        return

    # download the file
    response = requests.get(url, stream=True)
    if response.status_code == 200:
        with open(output_path, 'wb') as f:
            for chunk in response.iter_content(chunk_size=8192):
                f.write(chunk)
        print(f"Downloaded {url} to {output_path}")
    else:
        raise Exception(f"Failed to download file: {response.status_code} - {response.text}")


def extract_info(data_dir, catalog = 'GWTC'):
    # create the sample directory if it does not exist
    os.makedirs(f"{data_dir}/samples", exist_ok=True)

    # get the event list from gwosc
    events = {}
    for event in get_all_event(catalog=catalog):
        event_name = event['name']
        version = event['version']
        if event_name not in events:
            events[event_name] = event
        elif version > events[event_name]['version']:
            events[event_name] = event
        else:
            print(f"Skipping {event_name} version {version} as it is not the latest version.")

    # get the information for each event and download the preferred PE samples
    for event_name, event in events.items():
        print(event_name, f"v{event['version']}")
        # urls = get_event_urls(f"{event_name}-v{event['version']}", sample_rate=16384)
        
        gwosc_preferred_pe = get_prefered_pe(event_name, event["detail_url"])
        if gwosc_preferred_pe:
            print(f"{gwosc_preferred_pe['name']} data url: {gwosc_preferred_pe['data_url']}")

            outfile = f"{data_dir}/samples/{event_name}_combined_PEDataRelease.hdf"
            if os.path.exists(outfile):
                print(f"File {outfile} already exists, skipping download.")
            else:
                download_file(gwosc_preferred_pe['data_url'], outfile)
        else:
            print(f"⚠️ Event: {event_name}, Version: {event['version']}, No preferred PE found.")

def main():
    parser = argparse.ArgumentParser(description="Download GWOSC data for PE Automator")
    parser.add_argument('--data_dir', '-d', type=str, default='.', help='Path to the data directory where the samples will be downloaded')
    parser.add_argument('--catalog', '-c', type=str, default='GWTC', help='Catalog to use for event extraction')
    args = parser.parse_args()

    # extract information and download the preferred PE samples
    extract_info(args.data_dir, catalog=args.catalog)


if __name__ == "__main__":
    main()


from pe_automator.postprocessing.jobs import read_results_catalog
import os
import tempfile
import shutil
import numpy as np
import os
import bilby
from bilby_pipe.utils import parse_args
from bilby_pipe.data_analysis import create_analysis_parser, DataAnalysisInput
from bilby_pipe.data_generation import create_generation_parser, DataGenerationInput
from bilby.core.sampler import get_sampler_class
import argparse
from concurrent.futures import ProcessPoolExecutor, as_completed
from tqdm import tqdm
from bilby.core.result import reweight
import json

results_dir = 'results'

def main(results_dir, original_loglikelihood_dir, new_loglikelihood_dir, save_to=None):
    results_catalog = read_results_catalog(results_dir)
    # wf_model_ready_to_process = ['IMRPhenomXPNR', 'IMRPhenomXPHM', 'IMRPhenomTPHM_lal']
    wf_model_ready_to_process = ['IMRPhenomXPNR', 'IMRPhenomXPHM', 'IMRPhenomTPHM_lal', 
                                    'IMRPhenomTHM', 'IMRPhenomXE', 'IMRPhenomXHM',
                                    'IMRPhenomTPHM', 'IMRPhenomTEHM']
    new_evidences = {}
    for eventname in tqdm(results_catalog, desc="Processing events"):
        new_evidences[eventname] = {}
        for run in tqdm(results_catalog[eventname], desc=f"Processing {eventname}", leave=False):
            is_od_run = run['run_label'].startswith('OD') and not run['run_label'].startswith('OD16k')
            is_target_approximant = run['approximant'] in wf_model_ready_to_process
            if run['approximant'] not in new_evidences[eventname]:
                new_evidences[eventname][run['approximant']] = {}
            if is_od_run and is_target_approximant:
                recomputed_likelihoods_org = results_catalog.get_recomputed_likelihood(original_loglikelihood_dir, eventname, run['approximant'], run['run_label'])
                recomputed_likelihoods_new = results_catalog.get_recomputed_likelihood(new_loglikelihood_dir, eventname, run['approximant'], run['run_label'])
                if recomputed_likelihoods_org is None or recomputed_likelihoods_new is None:
                    print(f"Recomputed likelihoods not found for {eventname}, {run['approximant']}, {run['run_label']}. Skipping.")
                    continue
                print(f"Recomputing evidence for {eventname}, {run['approximant']}, {run['run_label']}")
                result = results_catalog.get_result(eventname, run['approximant'], run['run_label'])
                previous_log_evidence = result.log_evidence
                result.posterior['log_likelihood'] = recomputed_likelihoods_org
                reweighted_result = reweight(result, new_likelihood=recomputed_likelihoods_new)
                print(f"Event: {eventname}, Approximant: {run['approximant']}, Run label: {run['run_label']}")
                print(f"    Previous evidence: {previous_log_evidence:.2f}")
                print(f"  Recomputed evidence: {reweighted_result.log_evidence:.2f}")
                new_evidences[eventname][run['approximant']][run['run_label']] = reweighted_result.log_evidence

    # save new evidences to a json file
    if save_to:
        with open(save_to, 'w') as f:
            json.dump(new_evidences, f)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Recompute evidences for GW events using new likelihoods.")
    parser.add_argument('--results_dir', '-r', type=str, default='results', help='Directory containing the results catalog')
    parser.add_argument('--original_loglikelihood_dir', '-o', type=str, required=True, help='Directory containing the original recomputed likelihood files')
    parser.add_argument('--new_loglikelihood_dir', '-n', type=str, required=True, help='Directory containing the new recomputed likelihood files')
    parser.add_argument('--save_to', '-s', type=str, default='new_evidences.json', help='Path to save the new evidences JSON file')
    args = parser.parse_args()

    main(args.results_dir, args.original_loglikelihood_dir, args.new_loglikelihood_dir, save_to=args.save_to)




import os, glob, json
import numpy as np


import warnings
warnings.filterwarnings("ignore", "Wswiglal-redir-stdio")
import bilby

import matplotlib.pyplot as plt

def gwecc_TEHM_comparison_v2(res_dir:str, gw_ecc_dir:str, event:str):
    """
    Compare eccentricity samples from a gw_eccentricity and a Bilby TEHM runs.
    
    Parameters
    ----------
    dat_file : str
        Path to the .dat file produced by gw_eccentricity.
    hdf5_file : str
        Path to the Bilby .hdf5 result file produced by our run.
    """ 
    
    # Read gw_eccentricity postprocessed file
    gw_ecc_files = glob.glob(gw_ecc_dir+f'{event}/*json')

    # Check if gw_ecc file exist
    if len(gw_ecc_files)==0:
        raise ValueError(f"No gw eccentricity file for {event}")

    elif len(gw_ecc_files)>1: 
        raise ValueError(f"Multiple gw eccentricity files  = {gw_ecc_files} for {event}")

    else:
        gw_ecc_file = gw_ecc_files[0]

    basename = os.path.basename(gw_ecc_file)
    string_event = basename.split('_result')[0]

    # Read the corresponding posterior file
    res_files = glob.glob(res_dir+f"{event}/{string_event}*")

    # Perform some checks on the file
    if len(res_files)==0:
        raise ValueError(f"No res file for {event}")

    elif len(res_files)>1: 
        raise ValueError(f"Multiple posterior files  = {res_files} for {event}")

    else:
        res_file = res_files[0]
        
    ##########################################################
    
    # Read Bilby result
    result = bilby.result.read_in_result(res_file)
    posterior = result.posterior
    
    # Read gw_eccentricity file
    
    print(f"gw_ecc_file = {gw_ecc_file}")
    with open(gw_ecc_file) as json_data:
        result_gw_ecc = json.load(json_data)
    
    len_pe_samples = len(posterior['total_mass'])
    print('Number of samples from PE: ', len_pe_samples)

    #df0 = df0.where(df0 >= 0, np.nan)
    #samples = df0['eccentricity'].dropna().tolist()
    egw = np.array(result_gw_ecc['ew22_posterior'])
    lgw = np.array(result_gw_ecc[ 'meanAno22_posterior'] )
                        
    idx_bo = np.where(egw !=-1)[0]
    egw = egw[idx_bo]
    lgw = lgw[idx_bo]
    
    len_egw = len(egw)
    print('Number of samples in gw_eccentricity: ', len_egw)

    frac_survived = len_egw/len_pe_samples
    frac_dropped = 1 - frac_survived

    if frac_survived < 0.75:
        print(f"Warning: {frac_dropped*100:.1f}% of samples were dropped, do not trust the gw_eccentricity posterior.")

    plt.hist(posterior['eccentricity'], bins=60, density=True, histtype='step', lw=1.5, color='tomato', label='Bilby run TEHM')
    plt.hist(egw, bins=60, density=True, histtype='step', lw=1.5, color='mediumseagreen', label='gw_eccentricity')
    plt.xlabel("$e$", fontsize=16)
    plt.legend()
    plt.show()
    
    plt.hist(posterior['mean_anomaly'], bins=60, density=True, histtype='step', lw=1.5, color='tomato', label='Bilby run TEHM')
    plt.hist(lgw, bins=60, density=True, histtype='step', lw=1.5, color='mediumseagreen', label='gw_eccentricity')
    plt.xlabel("$l$", fontsize=16)
    plt.legend()
    plt.show()
    
    return




# from pe_automator.types.ssh import SimpleSSHClient
# from pe_automator.actions.allocation import get_allocation_info
# from pe_automator.actions.remote import run_command


# def main(allocation, hostname, user):
    

#     with SimpleSSHClient(name=allocation, hostname=hostname, username=user) as ssh_client:
#         ssh_client.connect()

#         run_command(f"bash env_test.sh", ssh_client=ssh_client)


# if __name__ == "__main__":
#     main('AECT-2025-2-0025', 'picasso.scbi.uma.es', 'resh000615')


import configparser


class FlexibleKeyDict(dict):
    """
    A dictionary that allows flexible key access, replacing '-' with '_' and vice versa, to handle the different key formats
    used in bilby pipe
    """
    def get(self, key, default=None):
        if key in self:
            return super().get(key, default)

        # Try replacing - with _ and vice versa
        alt_key_1 = key.replace('-', '_')
        alt_key_2 = key.replace('_', '-')

        if alt_key_1 in self:
            return super().get(alt_key_1, default)
        elif alt_key_2 in self:
            return super().get(alt_key_2, default)

        return default


def parse_dict_from_string(dict_string):
    dict_string = dict_string.strip('{} \n')
    # Split into key-value pairs
    dict_items = [item for item in dict_string.split(',') if item.strip()]
    # Build the dictionary
    parsed_dict = {}
    for item in dict_items:
        if ':' in item:
            key, value = item.split(':', 1)
            key = key.strip().strip('"').strip("'")
            value = value.strip().strip('"').strip("'")
           
            parsed_dict[key] = value

    return parsed_dict


def parse_array_from_string(array_string):
    if array_string.startswith('['):
        output_arr = eval(array_string)
    else:
        output_arr = [array_string]
    
    output_arr = [item.strip("'") for item in output_arr]
    
    return output_arr


def clean_string_array(input_array):
    """
    Clean an array of strings by removing leading and trailing whitespace and quotes.
    
    :param input_array: List of strings to clean.
    :return: List of cleaned strings.
    """
    return [item.strip().strip('"').strip("'") for item in input_array if item.strip()]


def convert_dict_to_string(input_dict):
    """
    Convert a dictionary to a string representation suitable for configuration files.
    
    :param input_dict: Dictionary to convert.
    :return: String representation of the dictionary.
    """
    return '{' + ', '.join(f'{k}: {v}' for k, v in input_dict.items()) + '}'


def parse_config(config_file):
    """
    Parse the configuration file and return a FlexibleKeyDict.
    
    :param config_file: Path to the configuration file.
    :return: FlexibleKeyDict containing the parsed configuration.
    """
    config = configparser.ConfigParser()
    config.read(config_file)

    parsed_config = FlexibleKeyDict()

    for section in config.sections():
        for key, value in config.items(section):
            if value.startswith('{') and value.endswith('}'):
                value = parse_dict_from_string(value)
            elif value.startswith('[') and value.endswith(']'):
                value = parse_array_from_string(value)
            parsed_config[f"{section}-{key}"] = value

    return parsed_config


import gitlab
import re
import json
import os
from pe_automator.actions.gitlab import get_information_from_issue
from pe_automator.actions.allocation import get_allocation_info, get_env_path
from pe_automator.types.ssh import SimpleSSHClient
from pe_automator.actions.remote import run_command


def rescue_run(issue_number, data_path, private_token, walltime=""):
    """
    Restart a run by re-creating the environment and re-submitting the job.
    """
    print(f"Restarting run for issue number {issue_number}...")
    with open(os.path.join(data_path, 'project', 'project.json'), 'r') as f:
        project_config = json.load(f)
    print(f"Loaded project configuration: {project_config}")
    gitlab_url, gitlab_project = project_config['gitlab_url'], project_config['gitlab_project']

    gl = gitlab.Gitlab(gitlab_url, private_token=private_token)
    project = gl.projects.get(gitlab_project)
    issue = project.issues.get(issue_number)

    run_info = get_information_from_issue(issue)
    print(f"Run information: {run_info}")


    allocation = run_info['allocation']
    conda_env = run_info['conda_env']
    allocation_info = get_allocation_info(allocation, data_path=data_path)
    conda_activate_command = f"{get_env_path(allocation_info)}/{conda_env}/bin/activate"

    user = run_info['user']
    hostname = run_info['remote'].split('@')[1]
    allocation = run_info['allocation']
    allocation_info = get_allocation_info(allocation, data_path=data_path)
    match = re.match(r'^(.*?)/runs/', run_info['run_dir'])
    if match:
        event_path = match.group(1)
        run_path = run_info['run_dir'][len(event_path) + 1:]
        print(f"Event path: {event_path}, Run path: {run_path}")
    else:
        raise ValueError("Run directory does not match expected format.")

    with SimpleSSHClient(name=allocation, hostname=hostname, username=user) as ssh_client:
        submit_file = run_command(f'cd {event_path} && ls {run_path}/submit/slurm*_master.sh', ssh_client=ssh_client)
        submit_file = submit_file.strip()
        print(f"Found submit file: {submit_file}")

        if walltime:
            print(f"Updating walltime to {walltime}...")
            # sed -E -i.bak "/analysis_[^ ]+_par[0-9]+\.sh\)\)/s/--time=[^[:space:]]+/--time=23:50:00/g" submit_script.sh
            run_command(f'cd {event_path} && sed -E -i.bak "/analysis_[^ ]+_par[0-9]+\.sh\)\)/s/--time=[^[:space:]]+/--time={walltime}/g" {submit_file}', ssh_client=ssh_client)

        run_command(f'source {conda_activate_command} && cd {event_path} && sbatch {submit_file}', ssh_client=ssh_client)


from ..actions.update_config import update_config
from ..actions.upload_files import upload_files
from ..actions.remote import run_command, upload_files, extract_submit_command, extract_run_dir, extract_job_id
from ..actions.gitlab import create_gitlab_issue, check_access_to_gitlab, update_job_status, issue_exists, generate_issue_title
from ..actions.allocation import get_allocation_info, get_env_path
from ..utils.encryption import encrypted_name
from ..utils.check import get_dir_size_MB
from ..types.ssh import MultiSSHManager, SimpleSSHClient
from pe_automator._version import __version__
from jinja2 import Template
from datetime import datetime
import os
import shutil
import json
import click


def setup_run(eventname, data_path, run_label, 
              account, partition, qos, approximant, 
              user, conda_env, private_token, allocation, 
              npoint=1000, nact=50, naccept=60, maxmcmc=20000,
              memory=300, cpu=None, walltime='71:40:00',
              dry_run=False, parallel_bilby=False, comment='', distance_marginalization=None,
              minimum_frequency=None, waveform_minimum_frequency=None, waveform_reference_frequency=None,
              encrypt_label=False,
              priors=None, mode_array=None):
    """
    Setup the run environment for the PE Automator workflow.
    This function is called before the main workflow execution.
    """
    with open(os.path.join(data_path, 'project', 'project.json'), 'r') as f:
        project_config = json.load(f)
    print(f"Loaded project configuration: {project_config}")
    gitlab_url, gitlab_project = project_config['gitlab_url'], project_config['gitlab_project']

    if not dry_run:
        check_access_to_gitlab(gitlab_url=gitlab_url, gitlab_project=gitlab_project, private_token=private_token)

    # check if the size of data_path/framefiles is less than 100MB, if true, raise a warning with click
    if not os.path.exists(os.path.join(data_path, 'framefiles')):
        raise FileNotFoundError(f"Data path {os.path.join(data_path, 'framefiles')} does not exist.")
    data_size_MB = get_dir_size_MB(os.path.join(data_path, 'framefiles'))
    if data_size_MB < 100:
        click.confirm(f"The size of the data path {os.path.join(data_path, 'framefiles')} is {data_size_MB:.2f} MB, which is less than 100 MB. "
                       "This could mean the git-lfs is not correctly installed and the framefiles are not correctly pulled. Do you want to continue?", abort=True)        
    
    #### Remote information
    # get allocation information
    allocation_info = get_allocation_info(allocation, data_path)
    hostname = allocation_info['hostname']

    # remote project path
    remote_project_path = os.path.join(allocation_info['scratch'], user, 'uib_o4a_catalog')

    if allocation_info['cluster'] == 'MareNostrum5':
        slurm_extra_lines = f"account={account} partition={partition} qos={qos}"
    else:
        slurm_extra_lines = f"constraint=cal mem={int(memory)}gb"

    #### Local setup
    # define the label
    date = datetime.now().strftime("%Y%m%d")
    label = f"{eventname}_{approximant}_{run_label}_{date}"
    if encrypt_label:
        encrypted_label = encrypted_name(data_path, eventname, approximant, run_label, date)
    else:
        encrypted_label = label

    remote = f"{user}@{hostname}"

    conda_activate_command = f"{get_env_path(allocation_info)}/{conda_env}/bin/activate"

    # set up the configuration file
    config_file = os.path.join(data_path, "configs", f"{eventname}_config.ini")
    template_file = os.path.join(data_path, "templates", "bilby_config.tpl.ini")

    #### Main workflow
    # check if the run exists in git issue
    if not dry_run:
        print(f"Checking if the run already exists in GitLab issues for event {eventname} with approximant {approximant} and label {run_label}")
        if issue_exists(eventname, approximant, run_label, private_token=private_token, gitlab_url=gitlab_url, gitlab_project=gitlab_project):
            print(f"Issue {generate_issue_title(eventname, approximant, run_label)} already exists. Please change the run label if you want to create a new run.")
            print("Exiting setup_run.")
            return
    
    config, files_to_copy = update_config(config_file=config_file, 
                           template_file=template_file, 
                           label=encrypted_label,
                           approximant=approximant,
                           priors=priors,
                           eventname=eventname,
                           slurm_extra_lines=slurm_extra_lines,
                           scheduler_analysis_time=walltime,
                           memory=memory,
                           npoint=npoint,
                           naccept=naccept,
                           nact=nact,
                           maxmcmc=maxmcmc,
                           waveform_minimum_frequency=waveform_minimum_frequency,
                           waveform_reference_frequency=waveform_reference_frequency,
                           minimum_frequency=minimum_frequency,
                           npool=cpu if cpu else allocation_info['cpu_per_node'],
                           mode_array=mode_array,
                           dist_margin=distance_marginalization
                           )

    # create output directory
    outdir = os.path.join("runs", label)
    os.makedirs(outdir, exist_ok=True)

    # save the config content to file
    config_path = os.path.join(outdir, "config.ini")
    with open(config_path, 'w') as f:
        f.write(config)

    # copy necessary files to the output directory
    for file_list in files_to_copy.values():
        for file_path in file_list:
            source_path = os.path.join(data_path, file_path)
            if os.path.exists(source_path):
                # create the destination directory if it doesn't exist
                dest_dir = os.path.dirname(os.path.join(outdir, file_path))
                os.makedirs(dest_dir, exist_ok=True)
                # copy the file
                shutil.copy(source_path, os.path.join(outdir, file_path))
            else:
                raise FileNotFoundError(f"Source file {source_path} does not exist.")
    
    if dry_run:
        print("Dry run mode is enabled. No files will be uploaded to the remote server.")
        print(f"Output directory: {outdir}")
        print(f"Config file: {config_path}")
        return

    # zip the output directory
    shutil.make_archive(outdir, 'zip', outdir)

    # upload the output directory with ssh
    # create a folder ~/uib_o4a_catalog
    with SimpleSSHClient(name=allocation, hostname=hostname, username=user) as ssh_client:
        ssh_client.connect()
        print(f"Creating project directory on cluster: {remote_project_path}")
        run_command(f"mkdir -p {remote_project_path}", ssh_client=ssh_client)

        # upload the zip file to the remote server
        print(f"Uploading files to remote server: {remote_project_path}")
        upload_files(f"{outdir}.zip", remote_project_path, remote)

        # unzip the file on the remote server
        print(f"Unzipping files on remote server: {remote_project_path}/{os.path.basename(outdir)}.zip")
        run_command(f"unzip -o {remote_project_path}/{os.path.basename(outdir)}.zip -d {remote_project_path}/{label}", ssh_client=ssh_client)

        # remove the zip file from the remote server
        print(f"Removing zip file from remote server: {remote_project_path}/{os.path.basename(outdir)}.zip")
        run_command(f"rm {remote_project_path}/{os.path.basename(outdir)}.zip", ssh_client=ssh_client)

        # run conda activate parallel_bilby && cd ~/uib_o4a_catalog/{label} && bilby_pipe config.ini
        print(f"Running bilby_pipe on remote server: {remote_project_path}/{label}")
        output = run_command(f"source {conda_activate_command} && cd {remote_project_path}/{label} && bilby_pipe config.ini", ssh_client=ssh_client)

        submit_command = extract_submit_command(output)

        print(f"Submitting job with command: {submit_command}")
        if allocation_info['cluster'] == 'MareNostrum5':
            submit_file = submit_command.strip().split()[1]
            preprocess_sub_file_command = f"sed -i -e '/^#SBATCH --mem=/d' -e 's/--mem=[^[:space:]]*//g' {submit_file} &&"
        else:
            preprocess_sub_file_command = ''
        job_info = run_command(f"source {conda_activate_command} && cd {remote_project_path}/{label} && {preprocess_sub_file_command} {submit_command}", ssh_client=ssh_client)

    job_id = extract_job_id(job_info)
    run_dir = extract_run_dir(submit_command)

    run_info = {
        "eventname": eventname,
        "approximant": approximant,
        "run_label": run_label,
        "label": label,
        "run_dir": os.path.join(remote_project_path, label, run_dir),
        "job_id": job_id,
        "remote": remote,
        "conda_env": conda_env,
        "priors": priors,
        "mode_array": mode_array,
        "distance_marginalization": distance_marginalization,
        "pipeline_version": __version__,
        "allocation": allocation,
        "user": user,
        "hostname": hostname,
        "npoint": npoint,
        "nact": nact,
        "naccept": naccept,
        "maxmcmc": maxmcmc,
        "memory": memory,
        "cpu": cpu if cpu else allocation_info['cpu_per_node'],
        "f_min": minimum_frequency,
        "waveform_f_min": waveform_minimum_frequency,
        "waveform_f_ref": waveform_reference_frequency,
        "comment": comment,
    }

    print(f"Run setup complete. Run info: {run_info}")

    # dump the run info to a JSON file
    run_info_file = os.path.join(outdir, "run_info.json")
    with open(run_info_file, 'w') as f:
        json.dump(run_info, f, indent=4)

    # create a GitLab issue with the run information
    issue = create_gitlab_issue(run_info, gitlab_url=gitlab_url, gitlab_project=gitlab_project, private_token=private_token)

    # update the job status in the GitLab issue
    update_job_status(issue, 'created')




# this is a command line script to extract the priors for GWTC-4 with given hdf5 files and save them to a file
import argparse
import h5py
import numpy as np
import re
import os
import shutil
import zipfile
import json
from math import floor, ceil

special_runs = {
    'GW230529_181500': 'C00:IMRPhenomXPHM:LowSpinSecondary',
    'GW250114_082203': 'bilby-IMRPhenomXPHM-SpinTaylor_prod-reweighted'
}

default_run_name = ['C00:IMRPhenomXPHM-SpinTaylor', 'C01:IMRPhenomXPHM', 'IMRPhenomPv2_posterior', 'IMRPhenomPv2NRT_highSpin_posterior', 'C01:IMRPhenomXPHM:HighSpin', 'C01:IMRPhenomPv2_NRTidal:HighSpin']


class FlexibleKeyDict(dict):
    """
    A dictionary that allows flexible key access, replacing '-' with '_' and vice versa, to handle the different key formats
    used in bilby pipe
    """
    def get(self, key, default=None):
        if key in self:
            return super().get(key, default)

        # Try replacing - with _ and vice versa
        alt_key_1 = key.replace('-', '_')
        alt_key_2 = key.replace('_', '-')

        if alt_key_1 in self:
            return super().get(alt_key_1, default)
        elif alt_key_2 in self:
            return super().get(alt_key_2, default)

        return default
    

def find_preferred_wf_run(eventname, pe_samples):
    run_names = list(pe_samples.keys())
    if special_runs.get(eventname):
        return special_runs[eventname]
    
    for preferred_run in default_run_name:
        if preferred_run in run_names:
            return preferred_run
    return None


def get_gw_name_from_hdf5_filename(filename):
    match = re.search(r'GW\d{6}_\d{6}', filename)
    if match:
        return match.group(0)
    
    match = re.search(r'GW\d{6}', filename)
    if match:
        return match.group(0)
    
    raise ValueError(f"Filename {filename} does not match expected format for GW name.")


def parse_dict_from_string(dict_string):
    dict_string = dict_string.strip('{} \n')
    # Split into key-value pairs
    dict_items = [item for item in dict_string.split(',') if item.strip()]
    # Build the dictionary
    parsed_dict = {}
    for item in dict_items:
        if ':' in item:
            key, value = item.split(':', 1)
            key = key.strip().strip('"').strip("'")
            value = value.strip().strip('"').strip("'")
           
            parsed_dict[key] = value

    return parsed_dict


def parse_array_from_string(array_string):
    if array_string.startswith('['):
        output_arr = eval(array_string)
    else:
        output_arr = [array_string]
    
    output_arr = [item.strip("'") for item in output_arr]
    
    return output_arr


def extract_prior(pe_samples, run_name):
    prior_keys = pe_samples[run_name]['priors']['analytic'].keys()

    prior = {}
    for key in prior_keys:
        value = pe_samples[run_name]['priors']['analytic'][key][0]
        if isinstance(value, bytes):
            value = value.decode('utf-8')
        if value.startswith('UniformInComponentsChirpMass') or value.startswith('UniformSourceFrame') or value.startswith('UniformInComponentsMassRatio'):
            # print(f"⚠️ Detected complex prior {value} for key {key}, replacing with bilby.gw.prior.{value}")
            value = f'bilby.gw.prior.{value}'
        prior[key] = value

    return prior


def extract_config(pe_samples, run_name):
    config_keys = pe_samples[run_name]['config_file']['config'].keys()

    config = FlexibleKeyDict({})
    for key in config_keys:
        value = pe_samples[run_name]['config_file']['config'][key][0]
        if isinstance(value, bytes):
            value = value.decode('utf-8')
        config[key] = value

    return config


def extract_psds(pe_samples, run_name):
    psd_keys = pe_samples[run_name]['psds'].keys()

    psds = {}
    for key in psd_keys:
        psds[key] = np.array(pe_samples[run_name]['psds'][key])

    return psds


def extract_cal_env(pe_samples, run_name):
    cal_env_keys = pe_samples[run_name]['calibration_envelope'].keys()

    cal_env = {}
    for key in cal_env_keys:
        cal_env[key] = np.array(pe_samples[run_name]['calibration_envelope'][key])

    return cal_env


def extract_info(hdf5_files, output_dir):
    os.chdir(output_dir)
    # create a tmp directory to store the priors
    os.makedirs('configs', exist_ok=True)
    os.makedirs('psds', exist_ok=True)
    os.makedirs('spline_cal_envs', exist_ok=True)
    os.makedirs('framefile_meta', exist_ok=True)

    for hdf5_file in hdf5_files:
        gw_name = get_gw_name_from_hdf5_filename(hdf5_file)
        try:
            f = h5py.File(hdf5_file, 'r')
        except Exception as e:
            print(f"⚠️ Failed to open {hdf5_file}: {e}")
            continue
        
        with f as pe_samples:
            wf_run_name = find_preferred_wf_run(gw_name, pe_samples)
            if wf_run_name is None:
                print(f"⚠️ No preferred wf run found for {gw_name} in {hdf5_file}, available runs: {list(pe_samples.keys())}")
                continue
            print(f"Extracting config for {gw_name} with run name {wf_run_name}")

            # check if it is compatible
            try:
                _ = pe_samples[wf_run_name]['config_file']['config']
            except:
                print(f"⚠️ Incompatible PE samples structure for {gw_name} in {hdf5_file} under run {wf_run_name}, skipping.")
                continue
            
            # extract the config for the given run name
            config = extract_config(pe_samples, wf_run_name)

            # rewrite channel dict
            channel_dict = parse_dict_from_string(config.get('channel-dict'))
            for key in channel_dict.keys():
                channel_dict[key] = f"GWOSC-16KHZ_R1_STRAIN"
            if 'channel-dict' in config:
                config['channel-dict'] = '{' + ', '.join(f'{k}: {v}' for k, v in channel_dict.items()) + '}'
            elif 'channel_dict' in config:
                config['channel_dict'] = '{' + ', '.join(f'{k}: {v}' for k, v in channel_dict.items()) + '}'
            else:
                config['channel-dict'] = '{' + ', '.join(f'{k}: {v}' for k, v in channel_dict.items()) + '}'

            # rewrite the prior
            if not config.get('prior-dict') or config.get('prior-dict') == 'None':
                print("⏳ Extracting prior...")
                prior = extract_prior(pe_samples, wf_run_name)
                prior_str = '{' + ', '.join(f"{k}: {v}" for k, v in prior.items()) + '}'
                config['prior-dict'] = prior_str
                config['prior-file'] = 'None'

            config['calibration-correction-type'] = 'template'

            with open(f'configs/{gw_name}_config.ini', 'wb') as f:
                for key, value in config.items():
                    f.write(f"{key} = {value}\n".encode('utf-8'))

            psd_dict = extract_psds(pe_samples, wf_run_name)
            cal_env_dict = extract_cal_env(pe_samples, wf_run_name)

            for ifo, psd in psd_dict.items():
                np.savetxt(f'psds/{gw_name}_{ifo}_psd.txt', psd)
            for ifo, cal_env in cal_env_dict.items():
                np.savetxt(f'spline_cal_envs/{gw_name}_{ifo}_spline_cal_env.txt', cal_env)

            detectors = parse_array_from_string(config['detectors'])
            trigger_time = float(config.get('trigger-time'))
            post_trigger_duration = float(config.get('post-trigger-duration', 0))
            duration = float(config['duration'])
            start_time = trigger_time + post_trigger_duration - duration
            end_time = start_time + duration
            required_data = {
                'gw_name': gw_name,
                'detectors': detectors,
                'channel_dict': channel_dict,
                'trigger_time': trigger_time,
                'post_trigger_duration': post_trigger_duration,
                'duration': duration,
                'start_time': floor(start_time),
                'end_time': ceil(end_time),
                'gwosc': True
            }
            # save the required data to a json file
            with open(f'framefile_meta/{gw_name}.json', 'w') as f:
                json.dump(required_data, f, indent=4)


def main():
    parser = argparse.ArgumentParser(description="Extract priors from GWOSCg HDF5 files.")
    parser.add_argument('gwosc_directory', type=str, help='Directory containing the GWOSC HDF5 files.')
    parser.add_argument('--output', type=str, default='data',
                        help='Output file to save the extracted priors (default: data)')
    

    args = parser.parse_args()

    gwosc_directory = os.path.abspath(args.gwosc_directory)
    args.output = os.path.abspath(args.output)

    # get all hdf5 files contains 'GW' in the gwosc_directory
    hdf5_files = [os.path.join(gwosc_directory, f) for f in os.listdir(gwosc_directory) if (f.endswith('.hdf5') or f.endswith('.h5') or f.endswith('.hdf')) and 'combined_PEDataRelease' in f]
    print(f"Found {len(hdf5_files)} HDF5 files in {gwosc_directory}")
    extract_info(hdf5_files, args.output)


if __name__ == "__main__":
    main()


from pe_automator.postprocessing.jobs import read_results_catalog
import os
import tempfile
import shutil
import numpy as np
import os
import bilby
from bilby_pipe.utils import parse_args
from bilby_pipe.data_analysis import create_analysis_parser, DataAnalysisInput
from bilby_pipe.data_generation import create_generation_parser, DataGenerationInput
from bilby.core.sampler import get_sampler_class
import argparse
from concurrent.futures import ProcessPoolExecutor, as_completed
from tqdm import tqdm


def generate_bilby_dump(config):
    args, unknown_args = parse_args([config], create_generation_parser())
    data = DataGenerationInput(args, unknown_args)
    data.save_data_dump()
    dump_file = os.path.join(data.data_directory, "_".join([data.label, "data_dump.pickle"]))
    print(f'Data dump saved to {dump_file}')
    return dump_file


def recompute_log_likelihood(bilby_dump_file, config, result_file, save_to=None):
    arg_string = f'{config} --data-dump-file {bilby_dump_file} --label bilby_mannual --sampling-seed 1 --sampler dynesty'

    args, unknown_args = parse_args(
            arg_string.split(),
            create_analysis_parser(usage=__doc__),
        )

    analysis = DataAnalysisInput(args, unknown_args)

    likelihood, priors = analysis.get_likelihood_and_priors()

    priors.fill_priors(likelihood, default_priors_file=None)

    sampler_class = get_sampler_class(analysis.sampler)

    sampler = sampler_class(
        likelihood,
        priors=priors,
        outdir=analysis.outdir,
        label=analysis.label,
        injection_parameters=None,
        meta_data=analysis.meta_data,
        result_class=analysis.result_class,
        conversion_function=analysis.parameter_generation,
        **analysis.sampler_kwargs,
    )

    sampler._set_sampling_method()

    sampler.kwargs["live_points"] = sampler.get_initial_points_from_prior(
                        sampler.nlive
                    )
    # sampler.use_ratio = True

    result = bilby.result.read_in_result(filename=result_file)

    log_likelihood = []
    for index, row in result.posterior.iterrows():
        if index % 3000 == 0:
            print(f'Processing row {index}')
        theta = [row[key] for key in sampler._search_parameter_keys]
        log_l = sampler.log_likelihood(theta)
        log_likelihood.append(log_l)
        
    log_likelihood = np.array(log_likelihood)
    if save_to is not None:
        np.savetxt(save_to, log_likelihood)
        print(f'Log likelihoods saved to {save_to}')

    return log_likelihood


def recompute_likelihood_worker(eventname, run, config_file, data_dir, output_dir):
    # print(f"Worker started for {eventname} with {run['approximant']} and label {run['run_label']}")
    recomputed_likelihood_file = os.path.join(output_dir, f"{eventname}_{run['approximant']}_{run['run_label']}_recomputed_likelihood.dat")
    if recomputed_likelihood_file is None:
        raise ValueError(f"Could not determine recomputed likelihood file path for {eventname} with {run['approximant']} and label {run['run_label']}.")
    
    # test if the config file exists
    if not os.path.isfile(config_file):
        print(f"Config file {config_file} does not exist. Skipping.")
        return
    if os.path.isfile(recomputed_likelihood_file):
        print(f"Recomputed likelihood file {recomputed_likelihood_file} exists. Skipping.")
        return

    print(f"Processing {eventname} with {run['approximant']} and label {run['run_label']}")

    with tempfile.TemporaryDirectory() as tmpdirname:
        # copy the config file to the temporary directory
        tmp_config_file = os.path.join(tmpdirname, os.path.basename(config_file))
        # rewrite the line outdir= to outdir=outputs
        with open(config_file, 'r') as f:
            lines = f.readlines()
        with open(tmp_config_file, 'w') as f:
            for line in lines:
                if line.startswith('outdir='):
                    f.write('outdir=outputs\n')
                elif line.startswith('mode-array=') or line.startswith('mode_array='):
                    f.write('mode-array=[None]\n')
                else:
                    f.write(line)
        # find data_dir/psds/eventname_* and copy to tmpdirname
        os.mkdir(os.path.join(tmpdirname, 'psds'))
        os.mkdir(os.path.join(tmpdirname, 'framefiles'))
        os.mkdir(os.path.join(tmpdirname, 'spline_cal_envs'))
        psd_files = [os.path.join('psds', f) for f in os.listdir(os.path.join(data_dir, 'psds')) if f.startswith(eventname)]
        frame_files = [os.path.join('framefiles', f) for f in os.listdir(os.path.join(data_dir, 'framefiles')) if f.startswith(eventname)]
        spline_cal_env_files = [os.path.join('spline_cal_envs', f) for f in os.listdir(os.path.join(data_dir, 'spline_cal_envs')) if f.startswith(eventname)]
        files_to_copy = psd_files + frame_files + spline_cal_env_files
        for f in files_to_copy:
            source_path = os.path.join(data_dir, f)
            dest_path = os.path.join(tmpdirname, f)
            print(f"Copying {source_path} to {dest_path}")
            shutil.copy(source_path, dest_path)
        
        pwd = os.getcwd()
        os.chdir(tmpdirname)
        try:
            dump_file = generate_bilby_dump(tmp_config_file)
        except Exception as e:
            print(f"Error occurred while generating bilby dump: {e}")
            return
        finally:
            os.chdir(pwd)

        dump_file = os.path.join(tmpdirname, dump_file)
        if not os.path.isfile(dump_file):
            print(f"Dump file {dump_file} does not exist. Skipping recompute.")
            return

        recompute_log_likelihood(dump_file, tmp_config_file, run['file_path'], save_to=recomputed_likelihood_file)


def recompute_likelihood_all(results_dir, data_dir, output_dir, n_processes=1):
    if not os.path.isdir(output_dir):
        os.makedirs(output_dir)

    results_catalog = read_results_catalog(results_dir)
    # wf_model_ready_to_process = ['IMRPhenomXPNR', 'IMRPhenomXPHM', 'IMRPhenomTPHM_lal']
    wf_model_ready_to_process = ['IMRPhenomXPNR', 'IMRPhenomXPHM', 'IMRPhenomTPHM_lal', 
                                 'IMRPhenomTHM', 'IMRPhenomXE', 'IMRPhenomXHM',
                                 'IMRPhenomTPHM', 'IMRPhenomTEHM']

    futures = []
    with ProcessPoolExecutor(max_workers=n_processes) as executor:
        for eventname in results_catalog:
            for run in results_catalog[eventname]:
                is_od_run = run['run_label'].startswith('OD') and not run['run_label'].startswith('OD16k')
                is_target_approximant = run['approximant'] in wf_model_ready_to_process
                if is_od_run and is_target_approximant:
                    config_file = results_catalog.get_config_path(eventname, run['approximant'], run['run_label'])
                    futures.append(
                        executor.submit(recompute_likelihood_worker, eventname, run, config_file, data_dir, output_dir)
                    )

        # progress bar over all submitted futures
        for future in tqdm(as_completed(futures), total=len(futures), desc="Processing events"):
            try:
                future.result()   # raises exception if worker failed
            except Exception as e:
                print(f"[MAIN] Worker failed: {e}")


def main():
    parser = argparse.ArgumentParser(description="Recompute likelihoods for GW events.")
    parser.add_argument('--results', '-r', type=str, required=True, help='Directory containing the results catalog.')
    parser.add_argument('--data', '-d', type=str, required=True, help='Directory containing the data files (psds, framefiles, spline_cal_envs).')
    parser.add_argument('--output', '-o', type=str, default='recomputed_likelihoods', help='Directory to save recomputed likelihood files.')
    parser.add_argument('--n-processes', '-n', type=int, default=1, help='Number of parallel processes to use.')
    args = parser.parse_args()

    recompute_likelihood_all(args.results, args.data, args.output, n_processes=args.n_processes)


if __name__ == "__main__":
    main()


import bilby 
import matplotlib.pyplot as plt 

import numpy as np  
import json
import os, math, re
import seaborn as sns 
import matplotlib.gridspec as gridspec 
import corner

import scipy
from scipy.stats import gaussian_kde

alkde  = 0.00
alhist = 0.75
histlw = 1.75
labfs  = 22

key_labels = {'mass_ratio': '$1/q$'
              , 'chirp_mass': r'$\mathcal{M}[M_{\odot}]$'
              , 'total_mass': r'$M[M_{\odot}]$'
              , 'mass_1': r'$m_1[M_{\odot}]$'
              , 'mass_2': r'$m_2[M_{\odot}]$'
              , 'mass_1_source': r'$m_1^{src}[M_{\odot}]$'
              , 'mass_2_source': r'$m_2^{src}[M_{\odot}]$'
              , 'total_mass_source':r'$M_T^{src}[M_{\odot}]$'
              , 'chi_eff': r'$\chi_\mathrm{eff}$'
              , 'chi_p': r'$\chi_\mathrm{p}$'
              , 'luminosity_distance': r'$d_L$[Mpc]'
              , 'dec': r'$\delta[rad]$'
              , 'ra': r'$\alpha[rad]$'
              , 'theta_jn': r'$\theta_\mathrm{JN}$[rad]'
              , 'a1':r'$a_1$'
              , 'a2':r'$a_2$'
              , 'a_1':r'$a_1$'
              , 'a_2':r'$a_2$'
              , 'geocent_time':r'$t_c$[s]'
              , 'tilt_1':r'$t_1$'
              , 'tilt_2':r'$t_2$'
              , 'chi_1':r'$\chi_1$'
              , 'chi_2':r'$\chi_2$'
              , 'eccentricity':r'$e_0$'
              , 'mean_anomaly':r'$\zeta_0$'
              ,'mean_per_ano':r'$\zeta_0$'
              , 'e_gw':r'$e_{\mathrm{gw}}$'
              , 'l_gw':r'$l_{\mathrm{gw}}$'
              ,'log_likelihood':'log_likelihood'
		      ,'final_spin_TPHM':r'$a_f$'
              ,'network_matched_filter_snr':r'SNR'
              }
sns.set_palette("deep")


# Remove some annoying panda warnings when plotting the posteriors
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)

# Matplotlib settings
import seaborn as sns
sns.set_palette("colorblind")
#sns.set(style="white",font_scale=2.)

golden = 1.6180339887498948482045868 
plt.rc('text', usetex=True) # Comment this out if you don't have LaTeX in your PATH!

#plt.style.use('test.mplstyle')

alkde  = 0.00
alhist = 0.75
histlw = 1.75
labfs  = 22

"""
key_labels = {'mass_ratio': '$1/q$'
              , 'chirp_mass': r'$\mathcal{M}[M_{\odot}]$'
              , 'total_mass': r'$M_T[M_{\odot}]$'
              , 'mass_1': r'$m_1[M_{\odot}]$'
              , 'mass_2': r'$m_2[M_{\odot}]$'
              , 'mass_1_source': r'$m_1^{src}[M_{\odot}]$'
              , 'mass_2_source': r'$m_2^{src}[M_{\odot}]$'
              , 'total_mass_source': r'$M_T^{src}[M_{\odot}]$'
              , 'chi_eff': r'$\chi_\mathrm{eff}$'
              , 'luminosity_distance': r'$d_L$[Mpc]'
              , 'redshift': r'$z$'
              , 'theta_jn': r'$\theta_\mathrm{JN}$[rad]'
              ,'eccentricity':r'$e_0$'
              ,'mean_anomaly':r'$\zeta_0$'
              ,'mean_per_ano':r'$\zeta_0$'
              ,'e_gw':r'$e_{\mathrm{gw}}$'
              ,'l_gw':r'$l_{\mathrm{gw}}$'
              }
""";

colors=['#377eb8', '#ff7f00', '#4daf4a',
                  '#f781bf', '#a65628', '#984ea3',
                  '#999999', '#e41a1c', '#dede00']

colors = ['#1E90FF','#DC143C','#7CFC00','#8A2BE2','#ff8f00','#002147'
          ,'#A3C1AD','#293133','#ff0038','#F0D077','#006699','#c90016','#0f4d92','#666666',
          '#50C878','#D80010','#7CEECE'
          ,'#1EB53A','#DA0A14','#00A1E4','#7BB135','#000000','#FF6C0C']
colors= ['#4C72B0', 
         '#DD8452' 
         , '#937860' 
    ,'#DA8BC3' 
    ,'#55A868' 
        ]

def read_inj_file(injection_file):
    """
    Function to read injection file used to create a frame file.
    """
    
    
    with open(injection_file) as json_file:
        injection_params= json.load(json_file)
        

    injection_params['total_mass']=injection_params['mass_1']+injection_params['mass_2']
    injection_params = bilby.gw.conversion.convert_to_lal_binary_black_hole_parameters(injection_params)[0]

    injection_params['chi_1'] = injection_params['a_1']*np.sin(injection_params['tilt_1'])*np.cos(injection_params['phi_12'])
    injection_params['chi_2'] = injection_params['a_2']*np.sin(injection_params['tilt_2'])*np.cos(injection_params['phi_12'])

    inj_params = bilby.gw.conversion.generate_all_bbh_parameters(injection_params)

    
    return inj_params




def generate_inj_dict(result):
    """
    Create the injection dictionary from the Bilby posterior file. It only works when the injection has been done using Bilby.
    """
        
    # Generate injection dictionary
    posterior_dict ={}
    inj_dict = {}
    for key in result.keys():
        posterior_dict[key] = result[key].posterior
        inj_params =  bilby.gw.conversion.convert_to_lal_binary_black_hole_parameters(result[key].injection_parameters)[0]
        inj_dict[key] = bilby.gw.conversion.generate_all_bbh_parameters(inj_params)


        if inj_dict[key]['mass_1']<inj_dict[key]['mass_2']:
            inj_dict[key]['mass_ratio'] = 1./inj_dict[key]['mass_ratio']

    return inj_dict 

def read_posteriors(files:list):

    basename_list = [os.path.basename(file0).split('.hdf5')[0].split('_merged')[0] for file0 in files]
    #basename_list = ['SEOBNRv5E']#, 'SEOBNRv5E old']
    tag_list = basename_list#[file0.split('q4_05_m01_e0_inj_')[1] for file0 in basename_list]

    tag_dict ={}
    result = {}
    result_plot_dict = {}
    for ii in range(len(basename_list)):
        file0 = files[ii]
        
        name =tag_list[ii]# basename_list[ii]
        tag_dict[name] = tag_list[ii]
        
        print(name)
        result[name] = bilby.read_in_result(file0)
        detectors = list (result[name].meta_data['likelihood']['interferometers'].keys())
        
        matched_filter_snr =0
        for det in detectors:
            matched_filter_snr += abs(result[name].posterior[det+'_matched_filter_snr'])**2.
        result[name].posterior['network_matched_filter_snr'] = np.sqrt( matched_filter_snr) 
        
        keys = list(result[name].posterior.keys())

        if 'mean_anomaly' not in keys and 'mean_per_ano' in keys:
            result[name].posterior['mean_anomaly'] = result[name].posterior['mean_per_ano']
        
        if 'mean_per_ano' not in keys and 'mean_anomaly' in keys:
            result[name].posterior['mean_per_ano'] = result[name].posterior['mean_anomaly']
            
        # m1 = result[name].posterior['mass_1']
        # m2 = result[name].posterior['mass_2']
        # m1_src = result[name].posterior['mass_1_source']
        # m2_src = result[name].posterior['mass_2_source']

        # result[name].posterior['mass_1'] = m2
        # result[name].posterior['mass_2'] = m1
        # result[name].posterior['mass_1_source'] = m2_src
        # result[name].posterior['mass_2_source'] = m1_src

        result_plot_dict[name] = result[name].posterior
        print('=====================================================')

    return result, result_plot_dict, tag_dict


# Functions to plot model-model injections
def posterior_comparison_inj(parameter,
                             result_dict,
                             outdir='./outdir/',
                             plot_injection=True,
                             label='test',
                             labels_legend_dict=None,
                             ylabel=True,
                            legend=True,
                            pos_legend=None,
                             ls_styles = None,
                             lw_styles = None,
                            legend_size=13,colors =  ['#4C72B0',  '#DD8452', '#937860','#DA8BC3' ,'#55A868'],
                             ylims= None, xlims= None,
                             inj_dict=None,
                             save_plot = False
                            ):
    """
    Function to plot 1D posteriors
    """
    

    posterior_dict ={}
    for key in result_dict.keys():
        posterior_dict[key] = result_dict[key].posterior 
    
    fig,ax = plt.subplots(figsize=(6,4),dpi=250)
    
     
    idx = 0
    for key in posterior_dict.keys():
        
        posterior = posterior_dict[key]
        pos_keys = posterior.keys()
        
        if ls_styles:
            ls = ls_styles[idx]
            lw = lw_styles[idx]
        else:
            ls = '-'
            lw=histlw
        if parameter in pos_keys:
            
            posterior = posterior_dict[key][parameter]
            Pr_Default_Lower  = np.percentile(posterior,10)
            Pr_Default_Upper  = np.percentile(posterior,90)
            best=np.median(posterior)
            lowerp=np.abs(Pr_Default_Lower-best)
            upperp=Pr_Default_Upper-best

            
            if labels_legend_dict:
                key_plot = labels_legend_dict[key]

            else:
                key_plot = key

            sns.histplot(data=posterior, 
                         element="step",
                         stat='density',
                         #bins=75,
                         color=colors[idx],
                         fill=False,
                         linewidth=histlw, alpha=alhist,label=key_plot)
            
            plt.axvline(Pr_Default_Lower, ls = '--', c = colors[idx], lw=lw)
            plt.axvline(Pr_Default_Upper, ls = '--', c = colors[idx], lw=lw)


        idx += 1
        
    if(plot_injection) and (inj_dict):
        plt.axvline(inj_dict[key][parameter], ls = '-', c = 'k', lw=2, 
                    label='Inj. value ')#+key)
    #else:
    #    plt.title(r"$%s^{+%s}_{-%s}$" % (np.round(best,2),np.round(lowerp,2),np.round(upperp,2)))
    
    plt.xlabel(key_labels[parameter],fontsize=20);
    if ylabel:
        plt.ylabel('Probability Density',fontsize=20);
    
    
    size=20
    ax.tick_params(axis='x', which='major', pad=8,width=1.5,length=2,  size=3, labelsize=size,direction='in')
    ax.tick_params(axis='x', which='minor', pad=8,width=1,  length=1.5,size=1.5, labelsize=size,direction='in')
    ax.tick_params(axis='y', which='major', pad=8,width=1.5,length=2,  size=3, labelsize=size,direction='in')
    ax.tick_params(axis='y', which='minor', pad=8,width=1., length=1.5,size=1.5, labelsize=size,direction='in')


    
    if legend :
        if pos_legend:
            plt.legend(loc = 'best',frameon=False, prop={'size': legend_size}, bbox_to_anchor=pos_legend,
                      handlelength=1);
        else:
            plt.legend(loc = 'best',frameon=False, prop={'size':legend_size},handlelength=1);            
    if ylims:
        ax.set_ylim(ylims[0],ylims[1])
    if xlims:
        ax.set_xlim(xlims[0],xlims[1])
         
    plt.tight_layout()

    

    if save_plot:
        if not os.path.exists(outdir):
            os.makedirs(outdir)
        plt.savefig(outdir+label+'_posterior_%s.png'%parameter, bbox_inches = 'tight')
        plt.savefig(outdir+label+'_posterior_%s.pdf'%parameter, bbox_inches = 'tight')


##### Functions to plot 2D posteriors
def truncate(n, decimals=0):
    multiplier = 10 ** decimals
    return math.ceil(n * multiplier) / multiplier

def round_limit(max_val):
    if max_val < 1:
        max_x_str = str(max_val)
        digits = len(re.search('\d+\.(0*)', max_x_str).group(1))
        return(truncate(max_val, decimals=digits+1))
    else:
        return(truncate(max_val, decimals = 0))

#-------------------------------------------
# a. Plot with 2D contour + x,y histogram
#-------------------------------------------
def triangle_plot_2d_axes(
    xbounds, ybounds, figsize=(8, 8),
    width_ratios=[4, 1], height_ratios=[1, 4], wspace=0.0, hspace=0.0,
    grid=False,high1d=1):
    """Initialize the axes for a 2d triangle plot.
    """
    high1d = high1d

    fig = plt.figure(figsize=figsize, dpi=200)
    gs = gridspec.GridSpec(
        2, 2,
        width_ratios=width_ratios, height_ratios=height_ratios,
        wspace=wspace, hspace=hspace)

    ax1 = plt.subplot(gs[0])
    ax3 = plt.subplot(gs[2])
    ax4 = plt.subplot(gs[3])

    ax1.minorticks_on()
    ax3.minorticks_on()
    ax4.minorticks_on()

    if grid:
        ax1.grid(which='major', ls='-')
        ax1.grid(which='minor', ls=':')
        ax3.grid(which='major', ls='-')
        ax3.grid(which='minor', ls=':')
        ax4.grid(which='major', ls='-')
        ax4.grid(which='minor', ls=':')

    # Get rid of tick labels
    ax1.xaxis.set_ticklabels([])
    ax4.yaxis.set_ticklabels([])

    # Use consistent x-axis and y-axis bounds in all 3 plots
    ax1.set_ylim(0, high1d)
    ax1.set_xlim(xbounds[0], xbounds[1])
    ax3.set_xlim(xbounds[0], xbounds[1])
    ax3.set_ylim(ybounds[0], ybounds[1])
    ax4.set_xlim(0, high1d)
    ax4.set_ylim(ybounds[0], ybounds[1])

    return fig, ax1, ax3, ax4


def create_fig_and_axes(xbounds, ybounds, figsize=(6, 6),high1d=1):
    fig, ax1, ax3, ax4 = triangle_plot_2d_axes(
        xbounds, ybounds, figsize=figsize, width_ratios=[4, 1],grid=False,
        height_ratios=[1, 4], wspace=0.0, hspace=0.0,high1d=high1d)
    return fig, ax1, ax3, ax4

#------------------------------------------------------------
# b. Plotting the main result: NRSur PHM (filled Contour)
#------------------------------------------------------------

def add_samples_to_fig_hist(name,samples_dict, parameter_1,parameter_2,xlims,ylims,color_dict, label_dict, alpha_dict, lw_dict, bins_dict,ax1,ax3,ax4, zorder=10,norm_factor_x=1,norm_factor_y=1,lwcontour=4, contourdict={}, inj_values = None):

    #name: name of pe result to plot
    #parameter 1 and parameter 2: parameters to plot in the 2D plot
    #zorder: if plottting multiple results, a larger zorder places the contour on top
    #norm_factor_x,y: parameter to scale the corner 1D histograms

    x = np.array(samples_dict[name][parameter_1])
    y = np.array(samples_dict[name][parameter_2])
    xlow, xhigh = xlims
    xsmooth = np.linspace(xlow, xhigh, 1000)
    ylow, yhigh = ylims
    ysmooth = np.linspace(ylow, yhigh, 1000)

    norm_factor_x=0.95*norm_factor_x
    norm_factor_y=0.95*norm_factor_y

    c = color_dict[name]
    label = label_dict[name]
    alpha = alpha_dict[name]
    lw = lw_dict[name]
    kde = gaussian_kde(x)
    max_val_x = max(kde(xsmooth))
    bb = bins_dict[name]

    #ax1.plot(xsmooth, norm_factor_x*kde(xsmooth), color=c, lw=lw, label=label,zorder=zorder)
    #ax1.fill_between(xsmooth, 0, norm_factor_x*kde(xsmooth), color=c, alpha=alpha,zorder=zorder)
    #ax1.hist(norm_factor_x*xsmooth, bins = 50, density = True, histtype='stepfilled', color=c, label=label)
    ax1.hist(x, bins=80,density=True,histtype='stepfilled',lw=lw, alpha=alpha,color=c)
    ax1.hist(x, bins=80,density=True,histtype='step',color=c,lw=lw,label=label)

    ax1.axvline(np.quantile(x, 0.05), color=c, ls='dashed')
    ax1.axvline(np.quantile(x, 0.95), color=c, ls='dashed')
    ax1.set_ylim((0,norm_factor_x*1.1))
    
    kde = gaussian_kde(y)
    max_val_y = max(kde(ysmooth))
    #ax4.plot(norm_factor_y*kde(ysmooth), ysmooth, color=c, lw=lw, label=label,zorder=zorder)
    #ax4.hist(norm_factor_y*ysmooth, bins = 50, histtype='stepfilled', color=c, label=label)
    ax4.hist(y, bins=80, orientation='horizontal',density=True,histtype='stepfilled',color=c,lw=lw, alpha=alpha)
    ax4.hist(y, bins=80, orientation='horizontal',density=True,histtype='step',color=c,lw=lw,label=label)
    #ax4.fill_betweenx(ysmooth, 0, norm_factor_y*kde(ysmooth), color=c, alpha=alpha,zorder=zorder)
    ax4.axhline(np.quantile(y, 0.05), color=c, ls='dashed')
    ax4.axhline(np.quantile(y, 0.95), color=c, ls='dashed')
    ax4.set_xlim((0,norm_factor_y*1.1))
    

    ax1.grid(False)
    ax4.grid(False)
    #the level=0.9 parameter, makes the 2D contours enclose the 90% of the posterior volume/surface.

    my_range = [[xlow, xhigh], [ylow, yhigh]]
    corner.hist2d(x, y, ax=ax3, range=my_range, color=c,
                  plot_datapoints=False, plot_density=True,smooth=True,#levels=(np.exp(-0.5),np.exp(-1)),
                  levels=[0.9],fill_contours=False,bins=bb, lw=lwcontour, contour_kwargs=contourdict
#                   fill_contours=False, no_fill_contours=False,
                  #contour_kwargs=dict(linewidth=4)
                 )


def add_samples_to_fig_hist_NORM(name,samples_dict, parameter_1,parameter_2,xlims,ylims,color_dict, label_dict, alpha_dict, lw_dict, bins_dict,ax1,ax3,ax4, zorder=10,norm_factor_x=1,norm_factor_y=1,lwcontour=4, contourdict={}, inj_values = None):

    #name: name of pe result to plot
    #parameter 1 and parameter 2: parameters to plot in the 2D plot
    #zorder: if plottting multiple results, a larger zorder places the contour on top
    #norm_factor_x,y: parameter to scale the corner 1D histograms

    x = np.array(samples_dict[name][parameter_1])
    y = np.array(samples_dict[name][parameter_2])
    xlow, xhigh = xlims
    xsmooth = np.linspace(xlow, xhigh, 1000)
    ylow, yhigh = ylims
    ysmooth = np.linspace(ylow, yhigh, 1000)

    norm_factor_x=0.95*norm_factor_x
    norm_factor_y=0.95*norm_factor_y

    c = color_dict[name]
    label = label_dict[name]
    alpha = alpha_dict[name]
    lw = lw_dict[name]
    kde = gaussian_kde(x)
    max_val_x = max(kde(xsmooth))
    bb = bins_dict[name]

    binsx = np.histogram_bin_edges(x, bins=80)
    idx = np.digitize(x, binsx) - 1  # Convert to 0-based index
    idx = np.clip(idx, 0, len(binsx) - 2)  # Avoid out-of-bounds indexing

    weightsx = np.ones_like(x) / np.diff(binsx)[idx]

    
    binsy= np.histogram_bin_edges(y, bins=80)
    idx = np.digitize(y, binsy) - 1  # Convert to 0-based index
    idx = np.clip(idx, 0, len(binsy) - 2)  # Avoid out-of-bounds indexing

    weightsy = np.ones_like(y) / np.diff(binsy)[idx]

    #ax1.plot(xsmooth, norm_factor_x*kde(xsmooth), color=c, lw=lw, label=label,zorder=zorder)
    #ax1.fill_between(xsmooth, 0, norm_factor_x*kde(xsmooth), color=c, alpha=alpha,zorder=zorder)
    #ax1.hist(norm_factor_x*xsmooth, bins = 50, density = True, histtype='stepfilled', color=c, label=label)
    ax1.hist(x, weights=weightsx,  bins=80,lw=lw, alpha=alpha,color=c)
    ax1.hist(x, weights=weightsx, bins=80,histtype='step',color=c,lw=lw,label=label)

    ax1.axvline(np.quantile(x, 0.05), color=c, ls='dashed')
    ax1.axvline(np.quantile(x, 0.95), color=c, ls='dashed')
    ax1.set_ylim((0,norm_factor_x*1.1))
    
    kde = gaussian_kde(y)
    max_val_y = max(kde(ysmooth))
    #ax4.plot(norm_factor_y*kde(ysmooth), ysmooth, color=c, lw=lw, label=label,zorder=zorder)
    #ax4.hist(norm_factor_y*ysmooth, bins = 50, histtype='stepfilled', color=c, label=label)
    ax4.hist(y, bins=80, weights=weightsy, orientation='horizontal',histtype='stepfilled',color=c,lw=lw, alpha=alpha)
    ax4.hist(y, bins=80, weights=weightsy, orientation='horizontal',histtype='step',color=c,lw=lw,label=label)
    #ax4.fill_betweenx(ysmooth, 0, norm_factor_y*kde(ysmooth), color=c, alpha=alpha,zorder=zorder)
    ax4.axhline(np.quantile(y, 0.05), color=c, ls='dashed')
    ax4.axhline(np.quantile(y, 0.95), color=c, ls='dashed')
    ax4.set_xlim((0,norm_factor_y*1.1))
    

    ax1.grid(False)
    ax4.grid(False)
    #the level=0.9 parameter, makes the 2D contours enclose the 90% of the posterior volume/surface.

    my_range = [[xlow, xhigh], [ylow, yhigh]]

    hist, xedges, yedges = np.histogram2d(x, y, bins=bb, density=True)

    # Normalize by bin area to match `density_norm="width"`
    bin_width_x = np.diff(xedges)[:, None]  # Column vector
    bin_width_y = np.diff(yedges)[None, :]  # Row vector
    hist *= (bin_width_x * bin_width_y)  # Scale densities by bin area

    # Compute bin indices and clip to avoid out-of-bounds errors
    idx_x = np.digitize(x, xedges) - 1
    idx_y = np.digitize(y, yedges) - 1

    idx_x = np.clip(idx_x, 0, len(xedges) - 2)  # Clip indices within range
    idx_y = np.clip(idx_y, 0, len(yedges) - 2)

    weights = np.ones_like(x) / (np.diff(xedges)[idx_x] * np.diff(yedges)[idx_y])

    corner.hist2d(
        x, y, ax=ax3, bins=bb, range=my_range, color=c, weights=weights,
        plot_datapoints=False, smooth=True,plot_density=True,
        levels=[0.9], fill_contours=False, lw=lwcontour, contour_kwargs=contourdict
    )

#----------------------------------------------------------------------------
# c. Plotting other models: Phenom PHM, SEOBNR PHM (transparent contours)
#-----------------------------------------------------------------------------

def add_samples_to_fig_nofilled_hist(name, samples_dict, parameter_1,parameter_2,xlims,ylims,color_dict, label_dict, alpha_dict, lw_dict, bins_dict,ax1,ax3,ax4, zorder=10,norm_factor_x=1,norm_factor_y=1, lwcontour=4, contourdict={}):

    x = np.array(samples_dict[name][parameter_1])
    y = np.array(samples_dict[name][parameter_2])
    xlow, xhigh = xlims
    xsmooth = np.linspace(xlow, xhigh, 1000)
    ylow, yhigh = ylims
    ysmooth = np.linspace(ylow, yhigh, 1000)

    norm_factor_x=0.95*norm_factor_x
    norm_factor_y=0.95*norm_factor_y

    c = color_dict[name]
    label = label_dict[name]
    alpha = alpha_dict[name]
    lw = lw_dict[name]
    kde = gaussian_kde(x)
    max_val_x = max(kde(xsmooth))
    bb = bins_dict[name]

    #ax1.plot(xsmooth, norm_factor_x*kde(xsmooth), color=c, lw=lw, label=label,zorder=zorder)
    #ax1.hist(norm_factor_x*xsmooth, bins = 50, density = True, histtype='step', color=c, label=label)
    ax1.hist(x, bins=80,density=True,histtype='step',color=c,lw=lw,label=label)

    ax1.axvline(np.quantile(x, 0.05), color=c, ls='dashed')
    ax1.axvline(np.quantile(x, 0.95), color=c, ls='dashed')
    ax1.set_ylim((0,norm_factor_x))

    kde = gaussian_kde(y)
    max_val_y = max(kde(ysmooth))

    #ax4.plot(norm_factor_y*kde(ysmooth), ysmooth, color=c, lw=lw, label=label,zorder=zorder)
    #ax4.hist(norm_factor_x*ysmooth, bins = 50, density = True, histtype='step', color=c, label=label)
    ax4.hist(y, bins=80, orientation='horizontal',density=True,histtype='step',color=c,lw=lw,label=label)
    ax4.axhline(np.quantile(y, 0.05), color=c, ls='dashed')
    ax4.axhline(np.quantile(y, 0.95), color=c, ls='dashed')
    ax4.set_xlim((0,norm_factor_y))

    my_range = [[xlow, xhigh], [ylow, yhigh]]
    corner.hist2d(x, y, ax=ax3, range=my_range, color=c,
                  plot_datapoints=False, plot_density=False,smooth=True,
                  levels=[0.9],no_fill_contour=False,bins=bb, lw=lwcontour, contour_kwargs=contourdict
                 )

def corner_plots_hist(result_posteriors, parameter_1, parameter_2, outdir,colors,no_solid=[],ncol=1, lw=1.5,lwcontour=4, alpha = 0.2, 
                      bins = 50, norm_factor = 1.60, save_plot=True, plotMaxL=False, loclegends='upper left', bboxlegend=(1.25, 1), xlims=None, 
                      ylims=None, contour_linewidth=2,legend_size = 12 ,xlabel=None, ylabel=None, fontsize = 24,
                      contour_linestyle='solid',showLegend=False,annotate='',annotate_position=(0.12,0.12),label_plot='GWXXXX', inj_dict = None, inj_dict_list = None):

    label_dict = {}
    color_dict = {}
    alpha_dict = {}
    lw_dict = {}
    bins_dict = {}
    contour_linewidth_dict = {}
    contour_linestyle_dict = {}
    if not isinstance(contour_linewidth,list):
        contour_linewidth = [contour_linewidth]*len(result_posteriors)
    if not isinstance(contour_linestyle,list):
        contour_linestyle = [contour_linestyle]*len(result_posteriors)

    for ind,key in enumerate(result_posteriors.keys()):
        label_dict[key]=key
        color_dict[key]=colors[ind]
        alpha_dict[key]=alpha
        bins_dict[key]=bins
        lw_dict[key]=lw
        contour_linewidth_dict[key] = contour_linewidth[ind]
        contour_linestyle_dict[key] = contour_linestyle[ind]

    xmin_list = []
    xmax_list = []
    ymin_list = []
    ymax_list = []

    for key in result_posteriors.keys():
        x = result_posteriors[key][parameter_1]
        y = result_posteriors[key][parameter_2]

        xmin = x.min()
        xmax = x.max()
        ymin = y.min()
        ymax = y.max()

        xmin_list.append(xmin)
        xmax_list.append(xmax)
        ymin_list.append(ymin)
        ymax_list.append(ymax)

    if xlims is None:
        xlims = [min(xmin_list),max(xmax_list)]
    if ylims is None:
        ylims = [min(ymin_list),max(ymax_list)]

    #if parameter_2 is 'log_likelihood':
    #    ylims = [0,max(ymax_list)+1]

    fig, ax1, ax3, ax4 = create_fig_and_axes(xlims, ylims)

    #ax3.set_xlabel(key_labels[parameter_1],fontsize=24)
    #ax3.set_ylabel(key_labels[parameter_2],fontsize=24)
    if xlabel:
        ax3.set_xlabel(xlabel,fontsize=fontsize)
    else:
        ax3.set_xlabel(key_labels[parameter_1],fontsize=fontsize)

    if ylabel:
        ax3.set_ylabel(ylabel,fontsize=fontsize)
    else:
        ax3.set_ylabel(key_labels[parameter_2],fontsize=fontsize)

    ax3.tick_params(labelsize=10)

    #ax3.set_xlabel(key_labels[parameter_1],fontsize=18)
    #ax3.set_ylabel(key_labels[parameter_2],fontsize=18)
    #ax3.tick_params(labelsize=8)#,direction='in')

    if parameter_1 == 'mass_ratio' or parameter_2 == 'mass_ratio':
        add = 0
    else:
        add = 1
    my_array_x=np.linspace(xlims[0]-add,xlims[1]+add,1000)
    my_array_y=np.linspace(ylims[0]-add,ylims[1]+add,1000)

    rx = []
    ry = []

    for key in result_posteriors.keys():

        xdat = result_posteriors[key][parameter_1]
        hhx,loc = np.histogram(xdat,density=True, bins=80)
        rxi = max(hhx)
        rx.append(rxi)

        ydat = result_posteriors[key][parameter_2]
        hhy,loc = np.histogram(ydat,density=True,bins = 80)
        ryi = max(hhy)
        ry.append(ryi)

    norm_factor_x = max(rx)
    norm_factor_y = max(ry)

    for key in result_posteriors.keys():

        if key in no_solid:
            add_samples_to_fig_nofilled_hist(key,result_posteriors,parameter_1,parameter_2,xlims,ylims,color_dict,
            label_dict, alpha_dict, lw_dict, bins_dict, ax1,ax3,ax4,zorder=20,
            norm_factor_x=round_limit(norm_factor_x),norm_factor_y=round_limit(norm_factor_y),
            lwcontour=lwcontour,contourdict={'linewidths':contour_linewidth_dict[key], 'linestyles':contour_linestyle_dict[key]})
        else:
            add_samples_to_fig_hist(key,result_posteriors,parameter_1,parameter_2,xlims,ylims,color_dict,
            label_dict, alpha_dict, lw_dict, bins_dict, ax1,ax3,ax4,zorder=20,
            norm_factor_x=round_limit(norm_factor_x),norm_factor_y=round_limit(norm_factor_y),
            lwcontour=lwcontour,contourdict={'linewidths':contour_linewidth_dict[key], 'linestyles':contour_linestyle_dict[key]})


    if inj_dict != None:
        ax1.axvline(x = inj_dict[parameter_1],ls = '-',color='gray',lw=lw)
        ax3.axvline(x = inj_dict[parameter_1],ls = '-',color='gray',lw=lw)
        ax4.axhline(y = inj_dict[parameter_2],ls = '-',color='gray',lw=lw)
        ax3.axhline(y = inj_dict[parameter_2],ls = '-',color='gray',lw=lw)
        
        ax3.scatter(inj_dict[parameter_1],inj_dict[parameter_2], s = 15,color='gray',lw=lw, marker='s')
    if inj_dict_list != None:

        for key in result_posteriors.keys():

            ax1.axvline(x = float(inj_dict_list[key][parameter_1]),ls = '-',color=color_dict[key],lw=lw)
            ax3.axvline(x = float(inj_dict_list[key][parameter_1]),ls = '-',color=color_dict[key],lw=lw)
            ax4.axhline(y = float(inj_dict_list[key][parameter_2]),ls = '-',color=color_dict[key],lw=lw)
            ax3.axhline(y = float(inj_dict_list[key][parameter_2]),ls = '-',color=color_dict[key],lw=lw)
            
            ax3.scatter(inj_dict_list[key][parameter_1],inj_dict_list[key][parameter_2], s = 15,color=color_dict[key], edgecolor='k',lw=lw, marker='X')


    if plotMaxL:
        for key in result_posteriors.keys():
            maxLpos = np.argmax(result_posteriors[key]['log_likelihood'].values)
            par1maxL = result_posteriors[key][parameter_1].values[maxLpos]
            par2maxL = result_posteriors[key][parameter_2].values[maxLpos]
            ax3.plot(par1maxL,par2maxL,marker='*', markersize=21, color=color_dict[key], markeredgecolor='black',markeredgewidth=1.0, linewidth=3)
    
    ax3.set_xlim(*xlims)
    ax1.set_yticklabels([],fontsize=10)
    ax4.set_xticklabels([],fontsize=20)
    if showLegend:
        ax3.legend(*ax4.get_legend_handles_labels(), loc=loclegends,bbox_to_anchor=bboxlegend, frameon=False,prop={'size': legend_size},ncol=ncol,handlelength=1)
    ax3.tick_params(axis='both', labelsize=20)
    
    ax3.annotate(annotate, xy=annotate_position, xycoords='axes fraction', fontsize = fontsize)#'x-large')
    if parameter_1 == 'mass_1_source' and parameter_2 == 'mass_2_source':

    
        xblack = [0,0,1000]
        yblack = [0,1000,1000]
        ax3.fill(xblack, yblack, 1000,facecolor='C7',zorder=10)

    
    ax3.grid(False)
    

    if not os.path.exists(outdir):
        os.makedirs(outdir)
    if save_plot:
        fig.savefig(outdir+'/cornerPlot_'+label_plot+'_'+parameter_1+'_'+parameter_2+'.png', transparent=True, bbox_inches='tight')
        #fig.savefig(outdir+'/cornerPlot_'+label_plot+'_'+parameter_1+'_'+parameter_2+'.pdf', transparent=True, bbox_inches='tight')



from bilby_pipe.bilbyargparser import BilbyConfigFileParser
from bilby_pipe.utils import convert_prior_string_input, convert_string_to_dict
from pe_automator.utils.parse_config import clean_string_array, convert_dict_to_string
from jinja2 import Template


def update_config(config_file, template_file, label, approximant, eventname, slurm_extra_lines, 
                  scheduler_analysis_time, memory, priors=None, mode_array=None, dist_margin=None,
                  minimum_frequency=None, waveform_minimum_frequency=None, waveform_reference_frequency=None,
                  npoint=1000, naccept=60, maxmcmc=20000, nact=50, npool=128):
    """
    Update the configuration file with the provided patch file.

    :param config_file: Path to the original configuration file.
    :param patch_file: Path to the patch file containing updates.
    """
    parser = BilbyConfigFileParser()

    with open(config_file, 'r') as f:
        config = parser.parse(f)[0]

    with open(template_file, 'r') as f:
        template = Template(f.read())

    detectors = clean_string_array(config.get("detectors"))

    PhenomXPY_Models = ['IMRPhenomTEHM', 'IMRPhenomTHM_20', 'IMRPhenomTHM_20mem', 'IMRPhenomTHM_20osc', 'IMRPhenomTHM', 'IMRPhenomTPHM', 'IMRPhenomTPHM_20', 'IMRPhenomXE', 'IMRPhenomXAS'] 
    ALIGN_SPIN_MODELS = ['IMRPhenomXHM_NSBH', 'IMRPhenomTEHM', 'IMRPhenomTHM_20', 'IMRPhenomTHM_20mem', 'IMRPhenomTHM_20osc', 'IMRPhenomTHM', 'IMRPhenomXE', 'IMRPhenomXHM', 'IMRPhenomXAS']
    extra_likelihood_kwargs = None
    if approximant == 'IMRPhenomXHM_NSBH':
        extra_likelihood_kwargs = {"highest_mode": 4, "linear_interpolation": True} 

    ################ waveform name setup ###############
    waveform_name_used = approximant
    if approximant in ['IMRPhenomTHM_20', 'IMRPhenomTHM_20mem', 'IMRPhenomTHM_20osc']:
        waveform_name_used = 'IMRPhenomTHM'
    if approximant in ['IMRPhenomTPHM_lal', 'IMRPhenomTPHM_20']:
        waveform_name_used = 'IMRPhenomTPHM'

    ################ waveform generator setup ###############
    waveform_generator = 'bilby.gw.waveform_generator.LALCBCWaveformGenerator'
    if approximant in ['IMRPhenomXE', 'IMRPhenomTEHM']:
        waveform_generator = 'bilby.gw.waveform_generator.WaveformGenerator'
    # if approximant == 'IMRPhenomXHM_NSBH' or approximant == 'IMRPhenomXPNR':
    #     waveform_generator = 'bilby.gw.waveform_generator.LALCBCWaveformGenerator'
    # elif approximant == 'IMRPhenomTEHM' or approximant == 'IMRPhenomTHM_20':
    #     waveform_generator = 'bilby.gw.waveform_generator.LALCBCWaveformGenerator'

    waveform_arguments_dict = None
    if approximant == 'IMRPhenomTEHM':
        waveform_arguments_dict = {'rhs_eqs':'eob_phenomT', 'force_condition':False}
    elif approximant == 'IMRPhenomTHM_20':
        waveform_arguments_dict = {'add_20_mode':True, 'contr_20':'full'}
    elif approximant == 'IMRPhenomTHM_20mem':
        waveform_arguments_dict = {'add_20_mode':True, 'contr_20':'memory'}
    elif approximant == 'IMRPhenomTHM_20osc':
        waveform_arguments_dict = {'add_20_mode':True, 'contr_20':'oscillatory'}
    elif approximant == 'IMRPhenomXE':
        waveform_arguments_dict = {'N_harmonics':2,'rhs_eqs':'eob_phenomT'}
    elif approximant == 'IMRPhenomXPHM':
        waveform_arguments_dict = {'PhenomXHMReleaseVersion': 122022, 'PhenomXPFinalSpinMod': 2, 'PhenomXPrecVersion': 320}
    elif approximant == 'IMRPhenomTPHM':
        waveform_arguments_dict = {'prec_version': "numerical"}
    elif approximant == 'IMRPhenomTPHM_20':
        waveform_arguments_dict = {'prec_version': "numerical", 'add_20_mode': True, 'contr_20': 'full', 'mem_l2': True, 'numba_rotation': "custom_modes"}

    frequency_domain_source_model = 'lal_binary_black_hole'
    if approximant == 'IMRPhenomXHM_NSBH':
        frequency_domain_source_model = 'lal_binary_neutron_star'
    # for phenomxpy models
    elif approximant in PhenomXPY_Models:
        frequency_domain_source_model = 'phenomxpy.common.bilby_frequency_phenomxpy'

    conversion_function = None
    generation_function = None
    if approximant in PhenomXPY_Models:
        conversion_function = 'bilby.gw.conversion.convert_to_lal_binary_black_hole_parameters'
        generation_function = 'bilby.gw.conversion.generate_all_bbh_parameters'

    ############### mode_array setup ###############
    if mode_array:
        mode_array_input = mode_array
    else:
        mode_array_input = [None]
        # if approximant in ['IMRPhenomTHM_20', 'IMRPhenomTHM_20mem', 'IMRPhenomTHM_20osc', 'IMRPhenomTPHM_20']:
        #     mode_array_input = [[2, 2], [2, -2], [2, 1], [2, -1], [3, 3], [3, -3], [4, 4], [4, -4], [5, 5], [5, -5], [2, 0]]
        # if approximant in ['IMRPhenomTEHM', 'IMRPhenomTHM']:
        #     mode_array_input = [[2, 2], [2, 1], [3, 3], [4, 4], [5, 5], [2, -2], [2, -1], [3, -3], [4, -4], [5, -5]]

    ################ frequencies ###############
    if config['minimum-frequency'].startswith('{'):
        minimum_frequency_dict = convert_string_to_dict(config['minimum-frequency'])
    else:
        config_minimum_frequency = float(config['minimum-frequency'])
        minimum_frequency_dict = {d: config_minimum_frequency for d in detectors}
        minimum_frequency_dict['waveform'] = config_minimum_frequency

    # update the minimum frequency for the waveform if it is provided
    if waveform_minimum_frequency is not None:
        minimum_frequency_dict['waveform'] = waveform_minimum_frequency
    
    # update the minimum frequency for each detector if it is provided
    if minimum_frequency is not None:
        for key in minimum_frequency_dict:
            if key != 'waveform':
                minimum_frequency_dict[key] = minimum_frequency

    reference_frequency = config['reference-frequency']
    if waveform_reference_frequency is not None:
        reference_frequency = waveform_reference_frequency

    ############### prior_dict setup ###############
    print(config["prior-dict"])
    prior_dict = convert_prior_string_input(config["prior-dict"])
    # for aligned spin models, replace the spin parameters with chi_1 and chi_2
    if approximant in ALIGN_SPIN_MODELS:
        keys_to_remove_for_aligned_spin = ["a_1", "a_2", "tilt_1", "tilt_2", "phi_12", "phi_jl", "spin_1x", "spin_1y", "spin_2x", "spin_2y"]
        for key in keys_to_remove_for_aligned_spin:
            if key in prior_dict:
                del prior_dict[key]
        prior_dict['chi_1'] = "bilby.gw.prior.AlignedSpin(name='chi_1', a_prior=Uniform(minimum=0, maximum=0.99))"
        prior_dict['chi_2'] = "bilby.gw.prior.AlignedSpin(name='chi_2', a_prior=Uniform(minimum=0, maximum=0.99))"

    # for the waveform approximants using phenomxpy, set the eccentricity prior to zero if the waveform model does not support eccentricity
    if approximant in ['IMRPhenomTHM_20', 'IMRPhenomTHM_20mem', 'IMRPhenomTHM_20osc', 'IMRPhenomTHM', 'IMRPhenomTPHM', 'IMRPhenomTPHM_20', 'IMRPhenomXAS']:
        prior_dict['eccentricity'] = "0"
        prior_dict['mean_anomaly'] = "0"

    # other changes required for specific approximants
    if approximant in ['IMRPhenomTEHM', 'IMRPhenomXE']:
        prior_dict['eccentricity'] = "Uniform(name='eccentricity', latex_label='$e$', minimum=0, maximum=0.65)"
        prior_dict['mean_anomaly'] = "Uniform(minimum=0, maximum=6.283185307179586, name='mean_anomaly', latex_label='$l_0$', unit=None, boundary='periodic')"
    
    if approximant == 'IMRPhenomXHM_NSBH':
        prior_dict['chirp_mass'] = "bilby.gw.prior.UniformInComponentsChirpMass(name='chirp_mass', minimum=1.45, maximum=9, latex_label='$\\mathcal{M}$', unit='$M_{\\odot}$', boundary=None)"
        prior_dict['mass_1'] = "Constraint(name='mass_1', minimum=3.0, maximum=50.0, latex_label='$m_{1}$', unit=None)"
        prior_dict['mass_2'] = "Constraint(name='mass_2', minimum=1.0, maximum=3.0, latex_label='$m_{2}$', unit=None)"
        prior_dict['chi_2'] = "bilby.gw.prior.AlignedSpin(a_prior=Uniform(name=None, minimum=0, maximum=0.05, latex_label=None, unit=None, boundary=None), z_prior=Uniform(name=None, minimum=-1, maximum=1, latex_label=None, unit=None, boundary=None), name='chi_2', latex_label='$\\chi_2$', unit=None, boundary=None)"
        prior_dict['luminosity_distance'] = "PowerLaw(alpha=2, name='luminosity_distance', minimum=10, maximum=1000, unit='Mpc')"
        prior_dict['lambda_1'] = "0"
        prior_dict['lambda_2'] = "Uniform(name='lambda_2', latex_label='$\lambda_2$', minimum=0, maximum=5000)"    

    # overwrite the default prior if a custom prior is provided
    if priors is not None:
        priors = convert_prior_string_input(priors)
        for key in priors:
            if key in prior_dict:
                prior_dict[key] = priors[key]
            else:
                print(f"Warning: Prior {key} not found in the default prior dictionary. Adding it.")
                prior_dict[key] = priors[key]

    default_prior = 'BBHPriorDict'
    if approximant in ['IMRPhenomXHM_NSBH']:
        default_prior = 'BNSPriorDict'

    ################ likelihood type setup ###############
    likelihood_type = config.get("likelihood-type", "GravitationalWaveTransient")
    if approximant in ['IMRPhenomXHM_NSBH']:
        likelihood_type = 'MBGravitationalWaveTransient'

    ################# marginalization setup ###############

    distance_marginalization = config['distance-marginalization']
    # if approximant in ['IMRPhenomTHM_20', 'IMRPhenomTHM_20mem', 'IMRPhenomTHM_20osc', 'IMRPhenomTPHM_20']:
    #     distance_marginalization = False

    if dist_margin is not None:
        distance_marginalization = dist_margin


    ################ local files ###############
    data_dict = {d: f'./framefiles/{eventname}_{d}.gwf' for d in detectors}
    psd_dict = {d: f'./psds/{eventname}_{d}_psd.txt' for d in detectors}
    spline_calibration_envelope_dict = {d: f'./spline_cal_envs/{eventname}_{d}_spline_cal_env.txt' for d in detectors}

    rendered = template.render(
        trigger_time=config.get("trigger-time"),
        detectors=detectors,
        data_dict=data_dict,
        channel_dict=config["channel-dict"],
        duration=config['duration'],
        post_trigger_duration=config['post-trigger-duration'],
        psd_dict=psd_dict,
        psd_fractional_overlap=config['psd-fractional-overlap'],
        psd_length=config['psd-length'],
        psd_maximum_duration=config['psd-maximum-duration'],
        psd_method=config['psd-method'],
        minimum_frequency=minimum_frequency_dict,
        maximum_frequency=config['maximum-frequency'],
        sampling_frequency=config['sampling-frequency'],
        tukey_roll_off=config['tukey-roll-off'],
        default_prior=default_prior,
        deltaT=config['deltaT'],
        prior_file=None,
        prior_dict=convert_dict_to_string(prior_dict),
        enforce_signal_duration=config.get('enforce-signal-duration', False),
        reference_frame=config['reference-frame'],
        time_reference=config['time-reference'],
        calibration_marginalization=config.get('calibration-marginalization', False),
        calibration_model=config['calibration-model'],
        calibration_prior_boundary=config.get('calibration-prior-boundary', 'reflective'),
        calibration_correction_type=config.get('calibration-correction-type', 'data'),
        spline_calibration_envelope_dict=spline_calibration_envelope_dict,
        spline_calibration_nodes=config['spline-calibration-nodes'],
        distance_marginalization=distance_marginalization,
        phase_marginalization=config['phase-marginalization'],
        time_marginalization=config['time-marginalization'],
        likelihood_type=likelihood_type,
        extra_likelihood_kwargs=extra_likelihood_kwargs,
        sampler_kwargs={'nlive': npoint, 'naccept': naccept, 'sample': 'acceptance-walk', 'walks': 100, 'npool': npool, 
                        'nact': nact, 'maxmcmc': maxmcmc, 
                        'check_point': True, 'check_point_plot': True, 
                        'check_point_delta_t': 3600, 'print_method': 'interval-60'},
        waveform_generator=waveform_generator,
        reference_frequency=reference_frequency,
        waveform_approximant=waveform_name_used,
        waveform_arguments_dict=waveform_arguments_dict,
        mode_array=mode_array_input,
        frequency_domain_source_model=frequency_domain_source_model,
        conversion_function=conversion_function,
        generation_function=generation_function,
        slurm_extra_lines=slurm_extra_lines,
        scheduler_analysis_time=scheduler_analysis_time,
        memory=memory,
        label=label,
        request_cpus=npool,
        outdir=f"runs/{label}",
    )

    files_to_copy = {
        'data_dict': [f for f in data_dict.values()],
        'psd_dict': [f for f in psd_dict.values()],
        'spline_calibration_envelope_dict': [f for f in spline_calibration_envelope_dict.values()],
    }

    return rendered, files_to_copy


import os
import datetime

def check_summary_exists(event_summary_dir: str) -> bool:
    # check if the summary file exists for the given event
    return os.path.isdir(event_summary_dir)


def check_if_posterior_exists(event_summary_dir: str, label: str) -> bool:
    # check if the posterior file exists for the given event
    event_posterior_file = os.path.join(event_summary_dir, "config", f"{label}_config.ini")
    return os.path.isfile(event_posterior_file)


def generate_label(run_info: dict) -> str:
    # Generate a label for the given run information
    label = f"{run_info['approximant']}_{run_info['run_label']}"
    return label


def get_summary_command(runs: list, labels: list, event_summary_dir: str) -> str:
    # Generate the summary command for the given event
    # summarypages --webdir plots/pe_summary/GW230709_122727 --existing_webdir plots/pe_summary/GW230709_122727 --add_to_existing --samples results/GW230709_122727/Sparrow_IMRPhenomTEHM_test_cat0_ext_prior_20250804_data0_1372940865-202158_analysis_H1L1_merge_result.hdf5 results/GW230709_122727/Sparrow_IMRPhenomXE_test_xe_1_20250814_data0_1372940865-202158_analysis_H1L1_merge_result.hdf5 results/GW230709_122727/Sparrow_IMRPhenomXPNR_test_cat0_1_20250728_data0_1372940865-202158_analysis_H1L1_merge_result.hdf5  --gw --multi_process 4 --disable_expert --disable_interactive --no_ligo_skymap --labels IMRPhenomTEHM_test_cat0 IMRPhenomXE_test_xe_1 IMRPhenomXPNR_test_cat0_1
    existing_option = ''
    if check_summary_exists(event_summary_dir):
        existing_option = f"--existing_webdir {event_summary_dir} --add_to_existing"
    command = f"summarypages --webdir {event_summary_dir} {existing_option} --samples {' '.join(runs)} --labels {' '.join(labels)} --gw --multi_process 4 --disable_expert --disable_interactive --no_ligo_skymap"
    return command


def generate_summary_command(eventname: str, event_entry: list, summary_folder: str) -> str:
    print(f"Generating summary command for event: {eventname}")
    # get the latest entry for each approximant
    approximants = set([run_info['approximant'] for run_info in event_entry])
    event_summary_dir = os.path.join(summary_folder, eventname)

    runs_tba = []
    labels_tba = []

    for approximant in approximants:
        latest_run = max(
            (run_info for run_info in event_entry if run_info['approximant'] == approximant),
            key=lambda x: datetime.datetime.fromisoformat(x['created_at'].replace("Z", "+00:00")),
            default=None
        )

        if not latest_run:
            continue

        label = generate_label(latest_run)

        if not check_if_posterior_exists(event_summary_dir, label):
            runs_tba.append(latest_run['file_path'])
            labels_tba.append(label)
        else:
            print(f"Posterior for {label} already exists, skipping.")

    if not runs_tba:
        return ''
    
    print(f"Runs to be added: {labels_tba}")

    summary_command = get_summary_command(runs_tba, labels_tba, event_summary_dir)
    return summary_command



import os
import glob
from pe_automator.actions.remote import download_files


def download_result_file(run, output_dir, user):
    """
    Download the result file for the given run.
    
    Parameters:
    - run: Dictionary containing run information including 'run_dir' and 'remote'.
    - output_dir: Directory where the result file will be downloaded.
    - user: User information for remote access.
    
    Returns:
    - str: Path to the downloaded result file.
    """
    remote = f"{user}@{run['hostname']}"
    result_file = f"{run['run_dir']}/result/*_merge_result.hdf5"

    output_dir = os.path.join(output_dir, run['eventname'])
    # Ensure the output directory exists
    os.makedirs(output_dir, exist_ok=True)
    
    # Download the result file
    download_files(remote_path=result_file, local_path=output_dir, remote=remote)


def download_corner_plots(run, output_dir, user):
    """
    Download the merged plots for the given run.
    
    Parameters:
    - run: Dictionary containing run information including 'run_dir' and 'remote'.
    - output_dir: Directory where the merged plots will be downloaded.
    - user: User information for remote access.
    
    Returns:
    - str: Path to the downloaded merged plots.
    """
    remote = f"{user}@{run['hostname']}"
    merged_plots = f"{run['run_dir']}/result/*_corner.png"
    
    output_dir = os.path.join(output_dir, run['eventname'])
    # Ensure the output directory exists
    os.makedirs(output_dir, exist_ok=True)
    
    # Download the merged plots
    download_files(remote_path=merged_plots, local_path=output_dir, remote=remote)


def download_config_file(run, output_dir, user):
    """
    Download the config file for the given run.
    
    Parameters:
    - run: Dictionary containing run information including 'run_dir' and 'remote'.
    - output_dir: Directory where the config file will be downloaded.
    - user: User information for remote access.
    
    Returns:
    - str: Path to the downloaded config file.
    """
    remote = f"{user}@{run['hostname']}"
    config_file = f"{run['run_dir']}/*_config_complete.ini"

    output_dir = os.path.join(output_dir, run['eventname'])
    # Ensure the output directory exists
    os.makedirs(output_dir, exist_ok=True)
    
    # Download the config file
    download_files(remote_path=config_file, local_path=output_dir, remote=remote)


def check_if_downloaded(run, output_dir):
    """
    Check if the result file and corner plots have been downloaded for the given run.
    
    Parameters:
    - run: Dictionary containing run information including 'eventname'.
    - output_dir: Directory where the results are expected to be downloaded.
    - user: User information for remote access.
    
    Returns:
    - bool: True if both files are downloaded, False otherwise.
    """
    label = run['run_dir'].split('/')[-1]
    result_file = os.path.join(output_dir, run['eventname'], f'{label}*_merge_result.hdf5')

    matched_files = glob.glob(result_file)
    if not matched_files:
        return False
    
    config_file = os.path.join(output_dir, run['eventname'], f'{label}*_config_complete.ini')
    matched_configs = glob.glob(config_file)
    if not matched_configs:
        return False

    return True


import os
import shutil


def upload_files(remote, remote_path, event_name, config, psds, calibration_envelopes, framefiles):
    """
    Upload PSDs, calibration envelopes, and framefiles to the specified destination.

    Args:
        remote (str): Remote server address.
        remote_path (str): Path on the remote server where files will be uploaded.
        event_name (str): Name of the event for which files are being uploaded.
        config (str): Path to the configuration file.
        psds (list): List of PSD file paths to upload.
        calibration_envelopes (list): List of calibration envelope file paths to upload.
        framefiles (list): List of framefile paths to upload.
    """
    # test with local as remote

    upload_file(remote, config, os.path.join(remote_path, event_name, "config"))
    for psd in psds:
        upload_file(remote, psd, os.path.join(remote_path, event_name, "psds"))
    for envelope in calibration_envelopes:
        upload_file(remote, envelope, os.path.join(remote_path, event_name, "calibration_envelopes"))
    for framefile in framefiles:
        upload_file(remote, framefile, os.path.join(remote_path, event_name, "framefiles"))


def upload_file(remote, file, remote_path):
    shutil.copy(file, remote_path)
    print(f"✅ Uploaded {file} to {remote_path}")
